
#include <string>

#include <isis/ivm/isasm.h>


// globals

extern "C" { 

int isasm_func_num = 0;
int instr_num = 0;

}

std::vector<std::string> IsasmNameTable;
std::vector<SYM_TABLE_ENTRY> SymbolTable;
std::vector<int> StringTable;
std::vector<std::pair<int,int> > IsasmLabelTable;
std::vector<std::pair<int,int> > LinenoTable;
std::vector<ISASM_FUNCTION> IsasmFunctionTable;

// functions

void reset_assembler_state(void)
{
    isasm_func_num = 0;
    instr_num = 0;
	ISASM_lineno = 1;

    IsasmNameTable.clear();
    SymbolTable.clear();
    StringTable.clear();
    IsasmLabelTable.clear();
    IsasmFunctionTable.clear();
}


int isasm_name_table_find(char * string)
{
    for (unsigned int i = 0; i < IsasmNameTable.size(); i++)
    {
        if (IsasmNameTable[i] == string)
            return i;
    }

    return -1;
}

int isasm_name_table_add(char * string)
{
	std::string stlString = string;

	size_t pos = stlString.find("\\n");
	while (pos != std::string::npos)
	{
		stlString.erase(pos,1);
		stlString.replace(pos,1,1,'\n');

		pos = stlString.find("\\n");
	}

	pos = stlString.find("\\t");
	while (pos != std::string::npos)
	{
		stlString.erase(pos,1);
		stlString.replace(pos,1,1,'\t');

		pos = stlString.find("\\t");
	}

    int i = isasm_name_table_find((char*)stlString.c_str());

    if (-1 == i)
    {
        IsasmNameTable.push_back(stlString);
        i = (int)IsasmNameTable.size()-1;
    }

    return i;
}

int isasm_name_table_contains(char * string)
{
    return (-1 != isasm_name_table_find(string));
}

const char * isasm_name_table_get(int name_index)
{
    if (name_index < 0 || name_index > (int)IsasmNameTable.size()-1)
        return NULL;
    
    return IsasmNameTable[name_index].c_str();
}

int string_table_find(int name_index)
{
    for (unsigned int i = 0; i < StringTable.size(); i++)
    {
        if (StringTable[i] == name_index)
            return i;
    }

    return -1;
}

int string_table_add(int name_index)
{
    int i = string_table_find(name_index);

    if (-1 == i)
    {
        StringTable.push_back(name_index);
        i = (int)StringTable.size()-1;
    }

    return i;
}

int string_table_contains(int name_index)
{
    return (-1 == string_table_find(name_index));
}

int symbol_table_find(int name_index, SYM_TYPE type)
{
    for (unsigned int i = 0; i < SymbolTable.size(); i++)
    {
        if (SymbolTable[i].name_index == name_index &&
            SymbolTable[i].type == type)
            return i;
    }

    return -1;
}

int symbol_table_find_global(int name_index)
{
    int i = symbol_table_find(name_index, SYM_GLOBAL_SCALAR);
    if (-1 != i) return i;
    return symbol_table_find(name_index, SYM_GLOBAL_ARRAY);
}

int symbol_table_find_local(int name_index)
{
    int i = symbol_table_find(name_index, SYM_LOCAL_SCALAR);
    if (-1 != i) return i;
    return symbol_table_find(name_index, SYM_LOCAL_ARRAY);
}

int symbol_table_find_param(int name_index)
{
    return symbol_table_find(name_index, SYM_PARAM_SCALAR);
}

int find_closest_symbol(int name_index)
{
    int index = symbol_table_find_local(name_index);
    if (-1 != index) return index;

    index = symbol_table_find_param(name_index);
    if (-1 != index) return index;

    return symbol_table_find_global(name_index);
}

int symbol_table_add(int name_index, SYM_TYPE type)
{
    int i = symbol_table_find(name_index, type);

    if (-1 == i)
    {
        SYM_TABLE_ENTRY entry;
        entry.name_index = name_index;
        entry.type = type;
        entry.extra_info = -1;

        SymbolTable.push_back(entry);
        i = (int)SymbolTable.size()-1;
    }

    return i;
}

int symbol_table_contains(int name_index, SYM_TYPE type)
{
    return (-1 != symbol_table_find(name_index, type));
}

void isasm_calc_func_locals_size(void)
{
    IsasmFunctionTable[isasm_func_num].locals_size = 0;

    for (int i = (int)SymbolTable.size()-1; 
         i >= 0 && SymbolTable[i].type != SYM_GLOBAL_SCALAR && 
            SymbolTable[i].type != SYM_GLOBAL_ARRAY; i--)
    {
        if (SymbolTable[i].type == SYM_LOCAL_SCALAR)
            IsasmFunctionTable[isasm_func_num].locals_size += 1;
        else if (SymbolTable[i].type == SYM_LOCAL_ARRAY)
            IsasmFunctionTable[isasm_func_num].locals_size += SymbolTable[i].extra_info;
    }
}

void isasm_clear_func_symbols(void)
{
    while (SymbolTable.size() > 0 && 
           SymbolTable.back().type != SYM_GLOBAL_SCALAR &&
           SymbolTable.back().type != SYM_GLOBAL_ARRAY)
        SymbolTable.pop_back();

    IsasmLabelTable.clear();
}

void set_symbol_extra_info(int symbol_index, int extra_info)
{
    if (symbol_index < 0 || symbol_index > (int)SymbolTable.size()-1)
        return;

    SymbolTable[symbol_index].extra_info = extra_info;
}

int count_global_size(void)
{
    int count = 0;

    for (unsigned int i = 0; i < SymbolTable.size(); i++)
    {
        if (SYM_GLOBAL_SCALAR == SymbolTable[i].type)
        {
            count++;
        }
        else if (SYM_GLOBAL_ARRAY == SymbolTable[i].type)
        {
            count += SymbolTable[i].extra_info;
        }
    }

    return count;
}

void lineno_table_add(int line_no, int instr_index)
{
    LinenoTable.push_back(std::pair<int,int>(line_no,instr_index));
}

int label_table_find(int name_index)
{
    for (unsigned int i = 0; i < IsasmLabelTable.size(); i++)
    {
        if (IsasmLabelTable[i].first == name_index)
            return i;
    }

    return -1;
}

int label_table_add(int name_index)
{
    int i = label_table_find(name_index);

    if (-1 == i)
    {
        IsasmLabelTable.push_back(std::pair<int,int>(name_index,-1));
        i = (int)IsasmLabelTable.size()-1;
    }

    return i;
}

void label_table_set(int label_index, int instr_index)
{
    if (label_index < 0 || label_index > (int)IsasmLabelTable.size()-1)
        return;

    IsasmLabelTable[label_index].second = instr_index;
}

int label_table_get(int label_index)
{
    if (label_index < 0 || label_index > (int)IsasmLabelTable.size()-1)
        return -1;

    return IsasmLabelTable[label_index].second;
}


int isasm_new_func(int name_index)
{
    ISASM_FUNCTION f;
    f.name_index = name_index;
    f.first_instr = instr_num;

    IsasmFunctionTable.push_back(f);

    return (int)IsasmFunctionTable.size()-1;
}

int isasm_find_func(int name_index)
{
    for (unsigned int i = 0; i < IsasmFunctionTable.size(); i++)
    {
        if (IsasmFunctionTable[i].name_index == name_index)
            return i;
    }

    return -1;
}

void isasm_add_param(int name_index)
{
    int sym_index = symbol_table_add(name_index,SYM_PARAM_SCALAR);
    IsasmFunctionTable[isasm_func_num].param_symbols.push_back(sym_index);
}

void isasm_add_local_scalar(int name_index)
{
    int sym_index = symbol_table_add(name_index,SYM_LOCAL_SCALAR);
    IsasmFunctionTable[isasm_func_num].local_symbols.push_back(sym_index);
}

void add_local_array(int name_index, int size)
{
    int sym_index = symbol_table_add(name_index,SYM_LOCAL_ARRAY);
    set_symbol_extra_info(sym_index,size);
    IsasmFunctionTable[isasm_func_num].local_symbols.push_back(sym_index);
}


