
#include <isis/ivm/isasm.h>

// globals

std::vector<unsigned char> InstructionTable;

std::vector<int> JumpTargets;
std::vector<int> CallTargets;
std::vector<int> CallHostTargets;
std::vector<int> CallHostNameIndicies;
std::vector<int> CallObjTargets;
std::vector<int> CallObjNameIndicies;

// functions

void reset_instructions_state(void)
{
    InstructionTable.clear();
    JumpTargets.clear();
    CallTargets.clear();
    CallHostTargets.clear();
    CallHostNameIndicies.clear();
    CallObjTargets.clear();
    CallObjNameIndicies.clear();
}

void add_instruction_bytes(unsigned char * bytes, int num_bytes)
{
    for (int i = 0; i < num_bytes; i++)
    {
        InstructionTable.push_back(bytes[i]);
    }
}

int get_next_instr_byte(void)
{
    return (int)InstructionTable.size();
}

int get_int_from_stream(int base_byte)
{
    if (base_byte < 0 || base_byte + 3 >= (int)InstructionTable.size())
        return -1;

    unsigned char b0 = InstructionTable[base_byte + 0];
    unsigned char b1 = InstructionTable[base_byte + 1];
    unsigned char b2 = InstructionTable[base_byte + 2];
    unsigned char b3 = InstructionTable[base_byte + 3];

    return  (b0 << 0) |
            (b1 << 8) |
            (b2 << 16) |
            (b3 << 24);
}

void set_int_to_stream(int base_byte, int value)
{
    if (base_byte < 0 || base_byte + 3 >= (int)InstructionTable.size())
        return;

    unsigned char b0 = 0xFF & (value >> 0);
    unsigned char b1 = 0xFF & (value >> 8);
    unsigned char b2 = 0xFF & (value >> 16);
    unsigned char b3 = 0xFF & (value >> 24);

    InstructionTable[base_byte + 0] = b0;
    InstructionTable[base_byte + 1] = b1;
    InstructionTable[base_byte + 2] = b2;
    InstructionTable[base_byte + 3] = b3;
}


void mark_jump_target(int offset)
{
    JumpTargets.push_back(get_next_instr_byte() + offset);
}

void resolve_label_jumps(void)
{
    for (unsigned int i = 0; i < JumpTargets.size(); i++)
    {
        int name_index = get_int_from_stream(JumpTargets[i]);

        int label_index = label_table_find(name_index);

        if (-1 == label_index)
            ISASM_quit("Unresolved jump target: %s\n",
                       isasm_name_table_get(name_index));

        int jump_instr = label_table_get(label_index);

        set_int_to_stream(JumpTargets[i],jump_instr);
    }

    JumpTargets.clear();
}


void mark_call_target(int offset)
{
    CallTargets.push_back(get_next_instr_byte() + offset);
}

void mark_callhost_target(int offset, int name_index)
{
    CallHostTargets.push_back(get_next_instr_byte() + offset);
    add_callhost_index(name_index);
}

void mark_callobj_target(int offset, int name_index)
{
    CallObjTargets.push_back(get_next_instr_byte() + offset);
    add_callobj_index(name_index);
}

int find_callhost_index(int name_index)
{
    for (unsigned int i = 0; i < CallHostNameIndicies.size(); i++)
    {
        if (CallHostNameIndicies[i] == name_index)
            return i;
    }

    return -1;
}

int find_callobj_index(int name_index)
{
    for (unsigned int i = 0; i < CallObjNameIndicies.size(); i++)
    {
        if (CallObjNameIndicies[i] == name_index)
            return i;
    }

    return -1;
}

void add_callhost_index(int name_index)
{
    if (-1 == find_callhost_index(name_index))
        CallHostNameIndicies.push_back(name_index);
}

void add_callobj_index(int name_index)
{
    if (-1 == find_callobj_index(name_index))
        CallObjNameIndicies.push_back(name_index);
}

void resolve_call_targets(void)
{
    for (unsigned int i = 0; i < CallTargets.size(); i++)
    {
        int name_index = get_int_from_stream(CallTargets[i]);

        int func_index = isasm_find_func(name_index);

        if (-1 == func_index)
            ISASM_quit("Unresolved function call: %s\n",
                       isasm_name_table_get(name_index));

        set_int_to_stream(CallTargets[i],func_index);
    }

    CallTargets.clear();
}

void resolve_callhost_targets(void)
{
    for (unsigned int i = 0; i < CallHostTargets.size(); i++)
    {
        int name_index = get_int_from_stream(CallHostTargets[i]);

        int callhost_index = find_callhost_index(name_index);

        if (-1 == callhost_index)
            ISASM_quit("Unresolved host function: %s\n",
                       isasm_name_table_get(name_index));

        set_int_to_stream(CallHostTargets[i],callhost_index);
    }
}

void resolve_callobj_targets(void)
{
    for (unsigned int i = 0; i < CallObjTargets.size(); i++)
    {
        int name_index = get_int_from_stream(CallObjTargets[i]);

        int callobj_index = find_callobj_index(name_index);

        if (-1 == callobj_index)
            ISASM_quit("Unresolved object function: %s\n",
                       isasm_name_table_get(name_index));

        set_int_to_stream(CallObjTargets[i],callobj_index);
    }
}


OPERAND_TYPE lval_op_type(LVAL * lval)
{
    switch (lval->type)
    {
        case LVAL_RETVAL:       return OP_RETVAL;
        case LVAL_ARRAY_LIT:
        case LVAL_SCALAR:       return OP_SCALAR;
        case LVAL_ARRAY_SYM:    return OP_ARRAY;
    }

    return (OPERAND_TYPE)-1;
}

OPERAND_TYPE rval_op_type(RVAL * rval)
{
    switch (rval->type)
    {
        case RVAL_LVAL:         return lval_op_type(&rval->val.lval);
        case RVAL_STRINGIND:    return OP_STRING;
        case RVAL_INTEGER:      return OP_INTEGER;
        case RVAL_FLOAT:        return OP_FLOAT;
    }

    return (OPERAND_TYPE)-1;
}


int calc_symbol_size(int symbol_num)
{
    switch(SymbolTable[symbol_num].type)
    {
    case SYM_GLOBAL_SCALAR:
    case SYM_LOCAL_SCALAR:
    case SYM_PARAM_SCALAR:
        return 1;
    case SYM_GLOBAL_ARRAY:
    case SYM_LOCAL_ARRAY:
        return SymbolTable[symbol_num].extra_info;
    }

    return -1;
}

int calc_stack_index(int symbol_num)
{
    switch(SymbolTable[symbol_num].type)
    {
    case SYM_GLOBAL_SCALAR:
    case SYM_GLOBAL_ARRAY:
        {
            int count = 0;

            for (int i = 0; i < symbol_num; i++)
                count += calc_symbol_size(i);

            return count;
        } break;
    case SYM_LOCAL_SCALAR:
    case SYM_LOCAL_ARRAY:
        {
            int count = 0;
            int start_symbol = 0;

            while ((SYM_LOCAL_SCALAR != SymbolTable[start_symbol].type) &&
                   (SYM_LOCAL_ARRAY != SymbolTable[start_symbol].type))
                start_symbol++;

            for (int i = start_symbol; i < symbol_num; i++)
                count += calc_symbol_size(i);

            return -((IsasmFunctionTable[isasm_func_num].locals_size-count-1)+3);
        } break;
    case SYM_PARAM_SCALAR:
        {
            int start_symbol = 0;

            while (SYM_PARAM_SCALAR != SymbolTable[start_symbol].type)
                start_symbol++;

            int count = IsasmFunctionTable[isasm_func_num].locals_size + 4 - 
                        (start_symbol - symbol_num);

            return -count;
        } break;
    }

    return -1;
}


void write_int(int val, unsigned char * dest, int * count)
{
    dest[0] = 0xFF & (val >> 0);
    dest[1] = 0xFF & (val >> 8);
    dest[2] = 0xFF & (val >> 16);
    dest[3] = 0xFF & (val >> 24);

    (*count) += 4;
}

void write_float(float val, unsigned char * dest, int * count)
{
    int i = *((int*)(&val));
    write_int(i,dest,count);
}

void write_lval(LVAL * lval, unsigned char * dest, int * count)
{
    int i = 0;
    int stack_index;

    dest[i++] = lval_op_type(lval);

    if (LVAL_SCALAR == lval->type)
    {
        stack_index = calc_stack_index(lval->symbol_index);
        write_int(stack_index,dest+i,&i);
    }
    else if (LVAL_ARRAY_LIT == lval->type)
    {
        stack_index = calc_stack_index(lval->symbol_index);
        write_int(stack_index + lval->offset.literal,dest+i,&i);
    }
    else if (LVAL_ARRAY_SYM == lval->type)
    {
        stack_index = calc_stack_index(lval->symbol_index);
        write_int(stack_index,dest+i,&i);
        write_int(calc_stack_index(lval->offset.symbol_index),dest+i,&i);
    }

    (*count) += i;
}

void write_rval(RVAL * rval, unsigned char * dest, int * count)
{
    int i = 0;
    dest[i++] = rval_op_type(rval);

    if (RVAL_STRINGIND == rval->type)
        write_int(rval->val.string_ind,dest+i,&i);
    else if (RVAL_INTEGER == rval->type)
        write_int(rval->val.int_val,dest+i,&i);
    else if (RVAL_FLOAT == rval->type)
        write_float(rval->val.float_val,dest+i,&i);
    else if (RVAL_LVAL == rval->type)
    {
        int stack_index = calc_stack_index(rval->val.lval.symbol_index);

        if (LVAL_SCALAR == rval->val.lval.type)
            write_int(stack_index,dest+i,&i);
        if (LVAL_ARRAY_LIT == rval->val.lval.type)
            write_int(stack_index + rval->val.lval.offset.literal,dest+i,&i);
        else if (LVAL_ARRAY_SYM == rval->val.lval.type)
        {
            write_int(stack_index,dest+i,&i);
            write_int(calc_stack_index(rval->val.lval.offset.symbol_index),
                      dest+i,&i);
        }
    }

    (*count) += i;
}

void emit_instr_lvalrval(INSTRUCTION_TYPE type, LVAL * lval, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = type;
    write_lval(lval, bytes + curr_byte,&curr_byte);
    write_rval(rval, bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_lval(INSTRUCTION_TYPE type, LVAL * lval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = type;
    write_lval(lval,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_rval(INSTRUCTION_TYPE type, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = type;
    write_rval(rval,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_lvalrvalrval(INSTRUCTION_TYPE type, LVAL * lval, RVAL * rval1,
                             RVAL * rval2)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = type;
    write_lval(lval,bytes + curr_byte,&curr_byte);
    write_rval(rval1,bytes + curr_byte,&curr_byte);
    write_rval(rval2,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_jmp(int name_index)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_JMP;

    mark_jump_target(curr_byte);
    write_int(name_index,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_jz(int name_index, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_JZ;
    write_rval(rval,bytes + curr_byte,&curr_byte);

    mark_jump_target(curr_byte);
    write_int(name_index,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_jnz(int name_index, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_JNZ;
    write_rval(rval,bytes + curr_byte,&curr_byte);

    mark_jump_target(curr_byte);
    write_int(name_index,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_call(int func_name_index, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_CALL;

    mark_call_target(curr_byte);
    
    write_int(func_name_index,bytes + curr_byte,&curr_byte);
    write_rval(rval,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_callhost(int func_name_index, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_CALLHOST;

    mark_callhost_target(curr_byte,func_name_index);

    write_int(func_name_index,bytes + curr_byte,&curr_byte);
    write_rval(rval,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_callobj(LVAL * object, int func_name_index, RVAL * rval)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_CALLOBJ;

    write_lval(object,bytes + curr_byte,&curr_byte);

    mark_callobj_target(curr_byte,func_name_index);
    write_int(func_name_index,bytes + curr_byte,&curr_byte);

    write_rval(rval,bytes + curr_byte,&curr_byte);

    add_instruction_bytes(bytes,curr_byte);
}

void emit_instr_this(void)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_THIS;

    add_instruction_bytes(bytes,curr_byte);
}


void emit_instr_ret(void)
{
    unsigned char bytes[100];
    int curr_byte = 0;

    bytes[curr_byte++] = INSTR_RET;

    add_instruction_bytes(bytes,curr_byte);
}

