
#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <isis/ivm/isc.h>


// globals

extern "C" {

int isc_func_num = 0;
int last_line = -1;

}

std::string g_sCurFile;

std::vector<std::string> IscNameTable;
std::vector<int> AssignedTable;
std::vector<int> LocalTable;
std::vector<ISC_FUNCTION> IscFunctionTable;
std::vector<std::pair<int,int> > GlobalTable;
std::vector<int> RegisterTable;
std::vector<int> IscLabelTable;
std::vector<LOOP_STRUCT> LoopStack;
std::vector<FOR_STRUCT> ForStack;

// functions

void reset_compiler_state(void)
{
    isc_func_num = 0;
	last_line = -1;
	ISC_lineno = 1;

    IscNameTable.clear();
    IscFunctionTable.clear();
    GlobalTable.clear();
    RegisterTable.clear();
    IscLabelTable.clear();
    LoopStack.clear();
    ForStack.clear();
}

void isc_clear_func_symbols()
{
    for (unsigned int i = 0; i < RegisterTable.size(); i++)
    {
        isc_add_local(isc_name_table_add(reg_name(i).c_str()),0);
    }

	RegisterTable.clear();
	IscLabelTable.clear();

    std::vector<int> WarningTable;

    for (unsigned int i = 0; i < LocalTable.size(); i++)
    {
        bool bFound = false;

        for (unsigned int j = 0; j < AssignedTable.size(); j++)
        {
            if (AssignedTable[j] == LocalTable[i])
                bFound = true;
        }

        if (!bFound)
            WarningTable.push_back(LocalTable[i]);
    }

    for (unsigned int i = 0; i < WarningTable.size(); i++)
    {
        if (std::string::npos == IscNameTable[WarningTable[i]].find("_r"))
        {
            printf("Warning: possibly unassigned variable \"%s\"\n"
                   "         in function %s\n"
                   "         in file %s\n",
                   isc_name_table_get(WarningTable[i]),
                   isc_name_table_get(IscFunctionTable.back().name_index),
                   g_sCurFile.c_str());
        }
    }

    AssignedTable.clear();
    LocalTable.clear();
}

void fixup_ism()
{
    unsigned int i;

    for (i = 0; i < IscFunctionTable.size(); i++)
    {
        if (IscFunctionTable[i].body.size() == 0 || 
            INSTR_RET != IscFunctionTable[i].body.back().instr)
        {
            LINE_STRUCT line;
            line.type = LINE_INSTR;
            line.instr = INSTR_MOV;
            line.ops.push_back("RetVal");
            line.ops.push_back("1");
            IscFunctionTable[i].body.push_back(line);

            line.instr = INSTR_RET;
            line.ops.clear();
            IscFunctionTable[i].body.push_back(line);
        }

        std::list<LINE_STRUCT>::iterator iter;
        for (iter = IscFunctionTable[i].body.begin();
             iter != IscFunctionTable[i].body.end();
             iter++)
        {
            if (LINE_INSTR == iter->type && INSTR_CALL == iter->instr)
            {
                bool found = false;

                for (unsigned int j = 0; j < IscFunctionTable.size(); j++)
                {
                    if (iter->ops[0] == 
                        isc_name_table_get(IscFunctionTable[j].name_index))
                    {
                        found = true;

                        if (atoi(iter->ops[1].c_str()) != 
                            (int)IscFunctionTable[j].param_symbols.size())
                        {
                            ISC_quit("Incorrect number of parameters for %s",
                           isc_name_table_get(IscFunctionTable[j].name_index));
                        }
                    }
                }

                if (!found)
                    iter->instr = INSTR_CALLHOST;
            }
        }
    }
}

int isc_name_table_find(const char * string)
{
    for (unsigned int i = 0; i < IscNameTable.size(); i++)
    {
        if (IscNameTable[i] == string)
            return i;
    }

    return -1;
}

int isc_name_table_add(const char * string)
{
    int i = isc_name_table_find(string);

    if (-1 == i)
    {
        IscNameTable.push_back(string);
        i = (int)IscNameTable.size()-1;
    }

    return i;
}

int isc_name_table_contains(const char * string)
{
    return (-1 != isc_name_table_find(string));
}

const char * isc_name_table_get(int name_index)
{
    if (name_index < 0 || name_index > (int)IscNameTable.size()-1)
        return NULL;
    
    return IscNameTable[name_index].c_str();
}

int isc_new_func(int name_index)
{
    ISC_FUNCTION f;
    f.name_index = name_index;

    IscFunctionTable.push_back(f);

    return (int)IscFunctionTable.size()-1;
}

int isc_find_func(int name_index)
{
    for (unsigned int i = 0; i < IscFunctionTable.size(); i++)
    {
        if (IscFunctionTable[i].name_index == name_index)
            return i;
    }

    return -1;
}

void isc_add_param(int name_index)
{
    IscFunctionTable[isc_func_num].param_symbols.push_back(name_index);
}

void isc_add_local(int name_index, int size)
{
    IscFunctionTable[isc_func_num].local_symbols.push_back(
                                          std::pair<int,int>(name_index,size));

    LocalTable.push_back(name_index);
}

void isc_add_assigned(int name_index)
{
    AssignedTable.push_back(name_index);
}

void add_global(int name_index, int size)
{
    GlobalTable.push_back(std::pair<int,int>(name_index,size));
}

int is_global_scalar(int name_index)
{
    for (unsigned int i = 0; i < GlobalTable.size(); i++)
    {
        if (GlobalTable[i].first == name_index &&
            GlobalTable[i].second == 0)
            return 1;
    }

    return 0;
}

int is_param(int name_index)
{
    for (unsigned int i = 0; 
         i < IscFunctionTable[isc_func_num].param_symbols.size(); 
         i++)
    {
        if (IscFunctionTable[isc_func_num].param_symbols[i] == name_index)
            return 1;
    }

    return 0;
}

int is_local_scalar(int name_index)
{
    for (unsigned int i = 0; 
         i < IscFunctionTable[isc_func_num].local_symbols.size(); 
         i++)
    {
        if (IscFunctionTable[isc_func_num].local_symbols[i].first == name_index && 
            IscFunctionTable[isc_func_num].local_symbols[i].second == 0)
            return 1;
    }

    return 0;
}

std::string instr_name(INSTRUCTION_TYPE type)
{
    switch (type)
    {
    case INSTR_MOV:      return "mov";

    case INSTR_ADD:      return "add";
    case INSTR_SUB:      return "sub";
    case INSTR_MUL:      return "mul";
    case INSTR_DIV:      return "div";
    case INSTR_MOD:      return "mod";
    case INSTR_INC:      return "inc";
    case INSTR_DEC:      return "dec";
    case INSTR_NEG:      return "neg";
    
    case INSTR_LOGICAND: return "logicand";
    case INSTR_LOGICOR:  return "logicor";
    case INSTR_BITAND:   return "bitand";
    case INSTR_BITOR:    return "bitor";
    case INSTR_XOR:      return "xor";
    case INSTR_NOT:      return "not";
    case INSTR_SHL:      return "shl";
    case INSTR_SHR:      return "shr";

    case INSTR_CONCAT:   return "concat";
    case INSTR_GETCHAR:  return "getchar";
    case INSTR_SETCHAR:  return "setchar";

    case INSTR_CMPE:     return "cmpe";
    case INSTR_CMPNE:    return "cmpne";
    case INSTR_CMPG:     return "cmpg";
    case INSTR_CMPGE:    return "cmpge";
    case INSTR_CMPL:     return "cmpl";
    case INSTR_CMPLE:    return "cmple";

    case INSTR_JMP:      return "jmp";
    case INSTR_JZ:       return "jz";
    case INSTR_JNZ:      return "jnz";

    case INSTR_PUSH:     return "push";
    case INSTR_POP:      return "pop";

    case INSTR_THIS:     return "this";

    case INSTR_CALL:     return "call";
    case INSTR_CALLHOST: return "callhost";
    case INSTR_CALLOBJ:  return "callobj";
    case INSTR_RET:      return "ret";
    }

    return "unknown";
}

void add_line(const LINE_STRUCT & line)
{
    if (ForStack.size() > 0 && ForStack.back().start)
    {
        ForStack.back().lines.push_back(line);
    }
    else
    {
        IscFunctionTable[isc_func_num].body.push_back(line);
    }
}

void add_instr(const LINE_STRUCT & line)
{
    if (last_line < ISC_lineno)
    {
        last_line = ISC_lineno;

        LINE_STRUCT no;
        no.type = LINE_LINENO;
        no.num = ISC_lineno;

        add_line(no);
    }

    add_line(line);
}

void emit_instr_0op(INSTRUCTION_TYPE type)
{
    LINE_STRUCT line;
    line.type = LINE_INSTR;
    line.instr = type;

    add_instr(line);
}

void emit_instr_1op(INSTRUCTION_TYPE type, const std::string & op0)
{
    LINE_STRUCT line;
    line.type = LINE_INSTR;
    line.instr = type;
    line.ops.push_back(op0);

    add_instr(line);
}

void emit_instr_2op(INSTRUCTION_TYPE type, const std::string & op0,
                    const std::string & op1)
{
    LINE_STRUCT line;
    line.type = LINE_INSTR;
    line.instr = type;
    line.ops.push_back(op0);
    line.ops.push_back(op1);

    add_instr(line);
}

void emit_instr_3op(INSTRUCTION_TYPE type, const std::string & op0,
                    const std::string & op1, const std::string & op2)
{
    LINE_STRUCT line;
    line.type = LINE_INSTR;
    line.instr = type;
    line.ops.push_back(op0);
    line.ops.push_back(op1);
    line.ops.push_back(op2);

    add_instr(line);
}

std::string line2string(const LINE_STRUCT & line)
{
    std::string text;

    switch (line.type)
    {
    case LINE_INSTR:
        {
            int i;

            text = "    ";
            text += instr_name(line.instr);

            for (i = 0; i < (int)line.ops.size()-1; i++)
            {
                text += " ";
                while ((int)text.length() < 5+15*(i+1))  { text += " "; }

                text += line.ops[i];
                text += ",";
            }

            if (i < (int)line.ops.size())
            {
                text += " ";
                while ((int)text.length() < 5+15*(i+1))  { text += " "; }

                text += line.ops[i];
            }

            text += ";\n";
        } break;
    case LINE_LABEL:
        {
            text = label_name(line.num) + ":\n";
        } break;
    case LINE_LINENO:
        {
            text = lineno_name(line.num) + "\n";
        } break;
    }

    return text;
}

void emit_assignment(EXPR_STRUCT * lval, EXPR_STRUCT * expr)
{
    switch (lval->type)
    {
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
        {
            emit_instr_2op(INSTR_MOV,expr_text(lval),expr_text(expr));
        } break;
    default:
        {
            ISC_quit("attempt to assign to a non-lvalue");
        } break;
    }
}

void emit_opequals(INSTRUCTION_TYPE instr, EXPR_STRUCT * lval, 
                   EXPR_STRUCT * expr)
{
    switch (lval->type)
    {
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
        {
            emit_instr_3op(instr,expr_text(lval),expr_text(lval),
                           expr_text(expr));
        } break;
    default:
        {
            ISC_quit("attempt to assign to a non-lvalue");
        } break;
    }
}

void emit_inc(EXPR_STRUCT * lval)
{
    switch (lval->type)
    {
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
        {
            emit_instr_2op(INSTR_INC,expr_text(lval),expr_text(lval));
        } break;
    default:
        {
            ISC_quit("attempt to increment a non-lvalue");
        } break;
    }
}

void emit_dec(EXPR_STRUCT * lval)
{
    switch (lval->type)
    {
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
        {
            emit_instr_2op(INSTR_DEC,expr_text(lval),expr_text(lval));
        } break;
    default:
        {
            ISC_quit("attempt to decrement a non-lvalue");
        } break;
    }
}

void emit_push_expr(EXPR_STRUCT * expr)
{
    emit_instr_1op(INSTR_PUSH,expr_text(expr));
}

std::string expr_text(EXPR_STRUCT * expr)
{
    switch (expr->type)
    {
    case EXPR_REGISTER:
        {
            return reg_name(expr->val.reg_num);
        } break;
    case EXPR_RVAL_SCALAR:
        {
            return isc_name_table_get(expr->val.name);
        } break;
    case EXPR_RVAL_ARRAY_REG:
        {
            return reg_name(expr->val.reg_num) + 
                               "[" + reg_name(expr->extra.reg) + "]";
        } break;
    case EXPR_RVAL_ARRAY_INT:
        {
            std::stringstream s;
            s << expr->extra.integer;
            return std::string(isc_name_table_get(expr->val.name)) + 
                               "[" + s.str() + "]";
        } break;
    case EXPR_RVAL_ARRAY_FLOAT:
        {
            std::stringstream s;
            s << expr->extra.floating;
            return std::string(isc_name_table_get(expr->val.name)) + 
                               "[" + s.str() + "]";
        } break;
    case EXPR_RVAL_ARRAY_STRING:
        {
            return reg_name(expr->val.reg_num) + 
                     "[\"" + isc_name_table_get(expr->extra.stringind) + "\"]";
        } break;
    case EXPR_RVAL_ARRAY_SCL:
        {
            return std::string(isc_name_table_get(expr->val.name)) + 
                             "[" + isc_name_table_get(expr->extra.name) + "]";
        } break;
    case EXPR_LIT_INTEGER:
        {
            std::stringstream s;
            s << expr->val.integer;
            return s.str();
        } break;
    case EXPR_LIT_FLOAT:
        {
            std::stringstream s;
            s << expr->val.floating;
            return s.str();
        } break;
    case EXPR_LIT_STRING:
        {
            return std::string("\"") + 
                           isc_name_table_get(expr->val.string_ind) + "\"";
        } break;
    }

    return "UNKNOWN";
}

int emit_reg_expression(EXPR_STRUCT * expr)
{
    int reg_num;

    switch (expr->type)
    {
    case EXPR_REGISTER:
        {
            reg_num = expr->val.reg_num;
        } break;
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
    case EXPR_LIT_INTEGER:
    case EXPR_LIT_FLOAT:
    case EXPR_LIT_STRING:
        {
            reg_num = find_free_reg();
            emit_instr_2op(INSTR_MOV,reg_name(reg_num),expr_text(expr));
        } break;
    default:
        {
            reg_num = -1;
        } break;
    }

    return reg_num;
}

int emit_1op(INSTRUCTION_TYPE instr, EXPR_STRUCT * expr)
{
    int reg_num = find_free_reg();

    emit_instr_2op(instr,reg_name(reg_num),expr_text(expr));

    return reg_num;
}

int emit_2op(INSTRUCTION_TYPE instr, EXPR_STRUCT * expr1, EXPR_STRUCT * expr2)
{
    int reg_num = find_free_reg();

    emit_instr_3op(instr,reg_name(reg_num),expr_text(expr1),expr_text(expr2));

    return reg_num;
}

void emit_return(EXPR_STRUCT * expr)
{
    if (NULL != expr)
    {
        emit_instr_2op(INSTR_MOV,"RetVal",expr_text(expr));
    }
    else
    {
        emit_instr_2op(INSTR_MOV,"RetVal","1");
    }

    emit_instr_0op(INSTR_RET);
}

int emit_new_call(int name_index, int num_params)
{
    int reg_num = find_free_reg();

    std::stringstream s;
    s << num_params + 1;

   emit_instr_1op(INSTR_PUSH,std::string("\"") + isc_name_table_get(name_index) 
                                                + "\"");
    emit_instr_2op(INSTR_CALL,"New",s.str());
    emit_instr_2op(INSTR_MOV,reg_name(reg_num),"RetVal");

    return reg_num;
}

int emit_this_call(void)
{
    int reg_num = find_free_reg();

    emit_instr_0op(INSTR_THIS);
    emit_instr_2op(INSTR_MOV,reg_name(reg_num),"RetVal");

	return reg_num;
}

int emit_call(int name_index, int num_params)
{
    int reg_num;
    std::stringstream s;

    s << num_params;
    reg_num = find_free_reg();

    emit_instr_2op(INSTR_CALL,isc_name_table_get(name_index),s.str());
    emit_instr_2op(INSTR_MOV,reg_name(reg_num),"RetVal");

    return reg_num;
}

int emit_callobj(EXPR_STRUCT * lval, int name_index, int num_params)
{
    int reg_num;
    std::stringstream s;

    s << num_params;
    reg_num = find_free_reg();

    switch (lval->type)
    {
    case EXPR_REGISTER:
    case EXPR_RVAL_SCALAR:
    case EXPR_RVAL_ARRAY_INT:
    case EXPR_RVAL_ARRAY_FLOAT:
    case EXPR_RVAL_ARRAY_STRING:
    case EXPR_RVAL_ARRAY_SCL:
    case EXPR_RVAL_ARRAY_REG:
        {
            emit_instr_3op(INSTR_CALLOBJ,expr_text(lval),
                           isc_name_table_get(name_index),s.str());
        } break;
    default:
        {
            ISC_quit("error: invalid object type used to call a method");
        } break;
    }

    emit_instr_2op(INSTR_MOV,reg_name(reg_num),"RetVal");

    return reg_num;
}

std::string reg_name(int reg_num)
{
    std::stringstream s;
    s << reg_num;

    return std::string("_r") + s.str();
}

std::string label_name(int label_num)
{
    std::stringstream s;
    s << label_num;

    return std::string("_l") + s.str();
}

std::string lineno_name(int line_num)
{
    std::stringstream s;
    s << line_num;

    return std::string("(") + s.str() + ")";
}

int find_free_reg(void)
{
    for (unsigned int i = 0; i < RegisterTable.size(); i++)
    {
        if (0 == RegisterTable[i])
        {
            RegisterTable[i] = 1;
            return i;
        }
    }

    RegisterTable.push_back(1);
    return (int)RegisterTable.size()-1;
}

void done_expr(EXPR_STRUCT * expr)
{
    if (EXPR_REGISTER == expr->type)
    {
        done_reg(expr->val.reg_num);
    }
    else if (EXPR_RVAL_ARRAY_REG == expr->type)
    {
        done_reg(expr->val.reg_num);
        done_reg(expr->extra.reg);
    }
}

void done_reg(int reg_num)
{
    if (reg_num < 0 || reg_num > (int)RegisterTable.size()-1)
        ISC_quit("INTERNAL COMPILER ERROR: invalid register: %d",reg_num);

    RegisterTable[reg_num] = 0;
}

int make_label()
{
    IscLabelTable.push_back(1);
    return (int)IscLabelTable.size()-1;
}

void insert_label(int label_num)
{
    LINE_STRUCT line;
    line.type = LINE_LABEL;
    line.num = label_num;

    add_line(line);
}

void insert_jmp(int label_num)
{
    emit_instr_1op(INSTR_JMP,label_name(label_num));
}

void insert_jz(EXPR_STRUCT * expr, int label_num)
{
    emit_instr_2op(INSTR_JZ,expr_text(expr),label_name(label_num));
}

void insert_jnz(EXPR_STRUCT * expr, int label_num)
{
    emit_instr_2op(INSTR_JNZ,expr_text(expr),label_name(label_num));
}

void push_loop(LOOP_STRUCT * loop)
{
    LoopStack.push_back(*loop);
}

void pop_loop()
{
    LoopStack.pop_back();
}

int get_continue_target()
{
    if (LoopStack.size() <= 0)
        ISC_quit("break or continue found outside of a loop");
    
    return LoopStack.back().continue_label;
}

int get_break_target()
{
    if (LoopStack.size() <= 0)
        ISC_quit("break or continue found outside of a loop");

    return LoopStack.back().end_label;
}

void push_for()
{
    FOR_STRUCT f;
    f.start = true;

    ForStack.push_back(f);
}

void pop_for()
{
    std::vector<LINE_STRUCT> lines = ForStack.back().lines;
    ForStack.pop_back();

    for (unsigned int i = 0; i < lines.size(); i++)
    {
        add_line(lines[i]);
    }
}

void end_forstart()
{
    ForStack.back().start = false;
}

