
#include <iostream>
#include <sstream>
#include <string>

#include <isis/ivm/isc.h>

// declarations

int write_string(const std::string & text, FILE * outfile);

int write_header(const std::string & infile_name, 
                  const std::string & outfile_name, FILE * outfile);
int write_globals(FILE * outfile);
int write_functions(FILE * outfile);

// definitions

void isc_reset_global_state(void)
{
    reset_compiler_state();
}

int save_ism_file(const std::string & infile_name, 
                  const std::string & outfile_name, FILE * outfile)
{
    if (!write_header(infile_name,outfile_name,outfile))
    {
        fprintf(stderr,"error writing ism header\n");
        return 0;
    }

    if (!write_globals(outfile))
    {
        fprintf(stderr,"error writing globals\n");
        return 0;
    }

    if (!write_functions(outfile))
    {
        fprintf(stderr,"error writing globals\n");
        return 0;
    }

    return 1;
}

int write_header(const std::string & infile_name,
                 const std::string & outfile_name, FILE * outfile)
{
    std::string text;

    text  = "//////////////////////////////////////////////////////////////\n";
    text += "//\n";
    text += std::string("// ") + PACKAGE + " version " + VERSION + 
           " - IsisScript Compiler compiled ISM file\n";
    text += "//\n";
    text += "// Input file: " + infile_name + "\n";
    text += "//\n";
    text += "// Output file: " + outfile_name + "\n";
    text += "//\n";
    text += "//////////////////////////////////////////////////////////////\n";
    text += "\n\n";

    if (!write_string(text,outfile))
    {
        fprintf(stderr,"error writing header\n");
        return 0;
    }

    return 1;
}

int write_globals(FILE * outfile)
{
    std::string text;

    text  = "////// Globals //////\n\n";

    for (unsigned int i = 0; i < GlobalTable.size(); i++)
    {
        text += std::string("var ") + isc_name_table_get(GlobalTable[i].first);

        if (0 < GlobalTable[i].second)
        {
            std::stringstream s;
            s << GlobalTable[i].second;
            text += std::string("[") + s.str() + "]";
        }

        text += ";\n";
    }

    text += "\n\n";
    
    if (!write_string(text,outfile))
    {
        fprintf(stderr,"error writing globals\n");
        return 0;
    }

    return 1;
}

int write_functions(FILE * outfile)
{
    unsigned int i,j;
    int k;
    std::string text;

    text  = "////// Functions //////\n\n";

    if (!write_string(text,outfile))
    {
        fprintf(stderr,"error writing functions header\n");
        return 0;
    }

    for (i = 0; i < IscFunctionTable.size(); i++)
    {
        text = std::string("func ") + 
                isc_name_table_get(IscFunctionTable[i].name_index) + "\n{\n";

        for (k = (int)IscFunctionTable[i].param_symbols.size()-1; k >= 0; k--)
        {
            text += std::string("    param ") + 
                    isc_name_table_get(IscFunctionTable[i].param_symbols[k]) + ";\n";
        }

        if (IscFunctionTable[i].param_symbols.size() > 0)
            text += "\n";

        for (j = 0; j < IscFunctionTable[i].local_symbols.size(); j++)
        {
            text += std::string("    var ") + 
                    isc_name_table_get(IscFunctionTable[i].local_symbols[j].first);

            if (0 < IscFunctionTable[i].local_symbols[j].second)
            {
                std::stringstream s;
                s << IscFunctionTable[i].local_symbols[j].second;
                text += std::string("[") + s.str() + "]";
            }

            text += ";\n";
        }

        if (0 < j)
            text += "\n";

        if (!write_string(text,outfile))
        {
            fprintf(stderr,"error writing function header\n");
            return 0;
        }

        for (std::list<LINE_STRUCT>::iterator iter = 
                                                IscFunctionTable[i].body.begin();
             iter != IscFunctionTable[i].body.end();
             iter++)
        {
            if (!write_string(line2string(*iter),outfile))
            {
                fprintf(stderr,"error writing function line: %d\n",j);
                return 0;
            }
        }

        text = "}\n\n";

        if (!write_string(text,outfile))
        {
            fprintf(stderr,"error writing function end\n");
            return 0;
        }
    }

    return 1;
}

int write_string(const std::string & text, FILE * outfile)
{
	size_t bw;

    bw = fwrite(text.c_str(),sizeof(char),text.length(),outfile);

    if ((sizeof(char) * text.length()) != bw)
    {
        fprintf(stderr,"error writing string: %s\n",text.c_str());
        return 0;
    }

    return 1;
}

