
#ifndef ISIS_GAME_GRAPHICSENGINE_H_INCLUDED
#define ISIS_GAME_GRAPHICSENGINE_H_INCLUDED 1

#include <map>
#include <string>
#include <list>
#include <vector>

#include <isis/ivm.h>

struct SDL_Surface;

namespace Isis {

class Image2D;
class Texture;
class CellImage;

class VectorAnimation;
class VectorImage;

class Font2D;

class VRAMObject;

class GraphicsEngine : public IVM::IVMObject, 
					   public IVM::ObjectProvider
{
public:

    GraphicsEngine();
    virtual ~GraphicsEngine();

	void Destroy();

    virtual void IVMFillExportTable(std::vector<std::string> & table);
    virtual bool IVMContainsMethod(const std::string & funcname);
    virtual IVM::Value IVMCallMethod(const std::string & funcname, 
		const std::vector<IVM::Value> & parameters);

	virtual void FillObjectTable(std::vector<std::string> & table);
    virtual IVM::IVMObject * CreateObject(const std::string & objname,
                                   const std::vector<IVM::Value> & InitParams);

    void CreateDisplay();
    void ToggleFullscreen();

    void Present();
    void Clear(unsigned int uiColor = 0xFF000000);

	Texture * LoadIGZ(const std::string & sImageFilename, 
					  std::vector<Image2D*> & SubImages);
	Texture * LoadIGZ(const std::string & sImageFilename); 

	void LoadVGZ(const std::string & sImageFilename,
				 std::vector<VectorAnimation*> & animations);

	void BindTexture(Texture * pTexture);
	void RebindCurrentTexture();

	void SetColor(unsigned int color);

    void VRAMObjectCreated(VRAMObject * pObject);
    void VRAMObjectDestroyed(VRAMObject * pObject);

	Font2D * LoadFont(const std::string & sIGZFilename);

    void SetWidth(int iWidth);
    int GetWidth();
    void SetHeight(int iHeight);
    int GetHeight();
    void SetWindowed(bool bWindowed);
    bool GetWindowed();
	void AddImagePrefix(const std::string & sImagePrefix);
	void RemoveImagePrefix(const std::string & sImagePrefix);
	std::vector<std::string> GetImagePrefixes();
    void EnableTexMap(bool bEnable);
    bool IsTexMapEnabled();
    void SetOrtho(bool bOrtho);
    bool IsOrtho();
    void SetDepth(bool bDepth);
    bool IsDepth();
	void SetBlend(bool bBlend);
	bool IsBlend();

	unsigned int Color(unsigned char red,
					   unsigned char green,
					   unsigned char blue,
					   unsigned char alpha = 255);

    void DrawRectangle(int x, int y, 
					   int width, int height, 
					   unsigned int color);
    void FillRectangle(int x, int y,
					   int width, int height,
					   unsigned int color);
    void DrawLine(int x1, int y1,
				  int x2, int y2, 
				  unsigned int color);

#ifdef _DEBUG
	static int GetGlobalRefCount();
#endif // _DEBUG

private:

    enum FuncType
    {
        FUNC_CREATEDISPLAY,
        FUNC_TOGGLEFULLSCREEN,
        FUNC_PRESENT,
        FUNC_CLEAR,
		FUNC_LOADIGZ,
		FUNC_LOADVGZ,
		FUNC_BINDTEXTURE,
		FUNC_REBINDCURRENTTEXTURE,
		FUNC_LOADFONT,
        FUNC_SETWIDTH,
        FUNC_GETWIDTH,
        FUNC_SETHEIGHT,
        FUNC_GETHEIGHT,
        FUNC_SETWINDOWED,
        FUNC_GETWINDOWED,
		FUNC_ADDIMAGEPREFIX,
		FUNC_REMOVEIMAGEPREFIX,
		FUNC_GETIMAGEPREFIXES,
		FUNC_ISTEXMAPENABLED,
        FUNC_ENABLETEXMAP,
		FUNC_ISORTHO,
        FUNC_SETORTHO,
		FUNC_ISDEPTH,
        FUNC_SETDEPTH,
		FUNC_ISBLEND,
		FUNC_SETBLEND,
		FUNC_COLOR,
		FUNC_DRAWRECTANGLE,
		FUNC_FILLRECTANGLE,
		FUNC_DRAWLINE,
		FUNC_PUSHMATRIX,
		FUNC_POPMATRIX,
		FUNC_LOOKAT,
		FUNC_PERSPECTIVE,
		FUNC_ROTATE,
		FUNC_TRANSLATE,
		FUNC_CULLFRONT,
		FUNC_CULLBACK
    };

    static std::map<std::string,FuncType> m_ExportTable;
    static void FillExportTable();

    enum ObjectType
    {
        OBJECT_TEXTURE,
		OBJECT_CELLIMAGE,
		OBJECT_FONT
    };

    static std::map<std::string,ObjectType> m_ObjectTable;
    static void FillObjectTable();

	bool m_bDestroyed;

	int m_iWidth;
    int m_iHeight;
    bool m_bWindowed;

    bool m_bTexMapEnabled;
    bool m_bOrtho;
    bool m_bDepth;
	bool m_bBlend;
	
	std::vector<std::string> m_ImagePrefixes;
	
    std::list<VRAMObject *> m_VRAMList;
	Texture * m_pBoundTexture;

	unsigned int m_Color;
    
    SDL_Surface * m_pMainSurface;

#ifdef _DEBUG
	static int m_iGlobalRefCount;
#endif // _DEBUG

};

} // namespace Isis

#endif // ISIS_GAME_GRAPHICSENGINE_H_INCLUDED

