
#ifndef IVM_ISISVM_H_INCLUDED
#define IVM_ISISVM_H_INCLUDED

#include <string>
#include <list>
#include <vector>
#include <map>

#include <isis/ivm/IVMObject.h>

namespace IVM {

class Script;
class ISEReader;
class ObjectProvider;

struct ISEStruct;

typedef std::list<Script *> ScriptList;
typedef std::map<std::string, ObjectProvider *> ObjProviderTable;
typedef std::map<std::string, IVMObject *> HAPITable;

class IsisVM : public IVMObject
{
public:

    IsisVM();
    virtual ~IsisVM();

	void Destroy();

    void AddScriptPrefix(const std::string & sScriptPrefix);
    void RemoveScriptPrefix(const std::string & sScriptPrefix);

    Script * MakeScript(const std::string & sScriptFilename,
                        const std::vector<Value> & InitParams);

    ISEStruct * LoadISE(const std::string & filename);
	std::string NameOfISE(ISEStruct * pISE);

    void RegisterHAPIProvider(IVMObject * provider);
    void RemoveHAPIProvider(IVMObject * provider);
    void RemoveAllHAPIProviders();
    bool HAPIProviderExists(const std::string & funcname);
    IVMObject * FindHAPIProvider(const std::string & funcname);

    void RegisterObjectProvider(ObjectProvider * provider);
    void RemoveObjectProvider(ObjectProvider * provider);
    void RemoveAllObjectProviders();
    bool ObjectProviderExists(const std::string & objname);
    ObjectProvider * FindObjectProvider(const std::string & objname);

    IVMObject * NewObject(const std::string & name, 
                          const std::vector<Value> & InitParams);

	Value CallAPI(const std::string & sFuncname, 
		          const std::vector<Value> & Params);

    virtual void IVMFillExportTable(std::vector<std::string> & table);
    virtual bool IVMContainsMethod(const std::string & funcname);
    virtual IVM::Value IVMCallMethod(const std::string & funcname, 
                               const std::vector<Value> & parameters);

#ifdef _DEBUG
	static int GetGlobalRefCount();
#endif // _DEBUG

private:

    enum FuncType
    {
        FUNC_ADDSCRIPTPREFIX,
        FUNC_REMOVESCRIPTPREFIX,
        FUNC_MAKESCRIPT,
        FUNC_REGISTERHAPIPROVIDER,
        FUNC_REMOVEHAPIPROVIDER,
        FUNC_HAPIPROVIDEREXISTS,
        FUNC_FINDHAPIPROVIDER,
        FUNC_GETVM,
        FUNC_NEW
    };

    static std::map<std::string,FuncType> m_ExportTable;
    static void FillExportTable();
	
	bool m_bDestroyed;

    HAPITable m_HAPITable;
    ObjProviderTable m_ObjectProviderTable;

	std::vector<std::string> m_ScriptPrefixes;

    std::map<std::string,ISEStruct *> m_ISEStructs;

    void FillInstructionTable(ISEStruct * ise, ISEReader & reader, 
                              int num_instr);
    void FillStringTable(ISEStruct * ise, ISEReader & reader, 
                         int num_strings);
    void FillFunctionTable(ISEStruct * ise, ISEReader & reader, 
                           int num_functions);
    void FillHostAPITable(ISEStruct * ise, ISEReader & reader, 
                          int num_functions);
    void FillHostObjTable(ISEStruct * ise, ISEReader & reader, 
                          int num_functions);
    void FillDebugTable(ISEStruct * ise, ISEReader & reader, 
                          int num_items);
    void CheckHAPIProviders(ISEStruct * ise);

#ifdef _DEBUG
	static int m_iGlobalRefCount;
#endif // _DEBUG

};

} // namespace IVM

#endif // IVM_ISISVM_H_INCLUDED

