
#include <SDL/SDL_opengl.h>
#include <ctype.h>

#include <isis/game/Font2D.h>
#include <isis/game/GraphicsEngine.h>
#include <isis/game/Image2D.h>

namespace Isis {

#ifdef _DEBUG
int Font2D::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,Font2D::FuncType> Font2D::m_ExportTable;

Font2D::Font2D(GraphicsEngine * pGraphicsEngine)
{
    if (m_ExportTable.empty())
        FillExportTable();

	m_pGraphicsEngine = pGraphicsEngine;
	m_pGraphicsEngine->AddRef();

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

Font2D::~Font2D()
{
	for (unsigned int i = 0; i < m_Characters.size(); i++)
		m_Characters[i]->Release();

	if (m_pGraphicsEngine)
		m_pGraphicsEngine->Release();

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int Font2D::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void Font2D::Init(const std::vector<Image2D *> & vecCharacters)
{
	if ('~' - ' ' + 1 != vecCharacters.size())
		throw IsisException("Not enough images found for a font");

	m_Characters = vecCharacters;

	for (unsigned int i = 0; i < m_Characters.size(); i++)
		m_Characters[i]->AddRef();
}

void Font2D::Init(const std::string & sIGZFilename) 
{
	if (!m_Characters.empty())
		throw IsisException("Already loaded in Load");

	m_pGraphicsEngine->LoadIGZ(sIGZFilename,m_Characters);

	if ('~' - ' ' + 1 != m_Characters.size())
		throw IsisException("Not enough images found for a font");

	for (unsigned int i = 0; i < m_Characters.size(); i++)
		m_Characters[i]->AddRef();
}

void Font2D::DrawString(const std::string & sString, int x, int y,
						unsigned int modColor)
{
	DrawString(sString,x,y,
			   m_pGraphicsEngine->GetWidth() - x,
			   m_pGraphicsEngine->GetHeight() - y,
			   modColor);
}

void Font2D::DrawString(const std::string & sString, int x, int y,
						int iBoundWidth, int iBoundHeight,
						unsigned int modColor)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in DrawString()");

	int iCurX = 0;
	int iCurY = 0;

	int iTextHeight = m_Characters[0]->GetHeight();

	for (unsigned int i = 0; i < sString.length(); i++)
	{
		int iWordPos = i;
		int iWordChar = sString[iWordPos] - ' ';
		int iWordSize = 0;

		while (sString[iWordPos] == ' ' && iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		while (!isspace(sString[iWordPos]) &&
				iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		if (iCurX + iWordSize > iBoundWidth)
		{
			iCurX = 0;
			iCurY += iTextHeight;

			if (isspace(sString[i]))
				i++;

			if (i >= sString.length())
				break;
		}

		int CharNum = sString[i] - ' ';

		if (iCurY + iTextHeight > iBoundHeight)
			return;

		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			m_Characters[CharNum]->Draw(x+iCurX-2,y+iCurY,
									0,0,Image2D::None,modColor);

			iCurX += m_Characters[CharNum]->GetWidth()-4;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurX += 4*(m_Characters[0]->GetWidth()-4);
				} break;
			case '\n':
				{
					iCurX = 0;
					iCurY += iTextHeight;
				} break;
			default: 
				break;
			}
		}
	}
}

void Font2D::CalcTextSize(const std::string & sString, 
						  int & iWidth, int & iHeight)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in CalcTextSize()");

	bool bNewchar = false;

	int iCurWidth = 0;

	iWidth = 0;
	iHeight = 0;

	for (unsigned int i = 0; i < sString.length(); i++)
	{
		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			int CharNum = sString[i] - ' ';

			iCurWidth += m_Characters[CharNum]->GetWidth()-4;

			if (iCurWidth > iWidth)
				iWidth = iCurWidth;

			bNewchar = true;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurWidth += 4*(m_Characters[0]->GetWidth()-4);

					if (iCurWidth > iWidth)
						iWidth = iCurWidth;
				} break;
			case '\n':
				{
					iCurWidth = 0;
					iHeight += m_Characters[0]->GetHeight();
					bNewchar = false;
				} break;
			default: 
				break;
			}
		}
	}

	if (bNewchar)
	{
		iHeight += m_Characters[0]->GetHeight();
	}
}

int Font2D::CalcNumLines(const std::string & sString,
						  int iLineWidth)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in CalcNumLines()");

	int iCurX = 0;
	int iNumLines = 1;

	for (unsigned int i = 0; i < sString.length(); i++)
	{
		int iWordPos = i;
		int iWordChar = sString[iWordPos] - ' ';
		int iWordSize = 0;

		while (sString[iWordPos] == ' ' && iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		while (!isspace(sString[iWordPos]) &&
				iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		if (iCurX + iWordSize > iLineWidth)
		{
			iCurX = 0;
			iNumLines++;

			if (isspace(sString[i]))
				i++;

			if (i >= sString.length())
				break;
		}

		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			int CharNum = sString[i] - ' ';

			iCurX += m_Characters[CharNum]->GetWidth()-4;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurX += 4*(m_Characters[0]->GetWidth()-4);
				} break;
			case '\n':
				{
					iCurX = 0;
					iNumLines++;
				} break;
			default: 
				break;
			}
		}
	}

	return iNumLines;
}


std::string Font2D::ChopLines(const std::string & sString, int NumLines,
						  int iLineWidth)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in CalcNumLines()");

	int iCurX = 0;
	int iNumLines = 0;

	if (NumLines <= 0)
		return sString;

	for (unsigned int i = 0; i < sString.length(); i++)
	{
		int iWordPos = i;
		int iWordChar = sString[iWordPos] - ' ';
		int iWordSize = 0;

		while (sString[iWordPos] == ' ' && iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		while (!isspace(sString[iWordPos]) &&
				iWordPos < (int)sString.length())
		{
			iWordSize += m_Characters[iWordChar]->GetWidth()-4;
			iWordPos++;
			iWordChar = sString[iWordPos] - ' ';
		}

		if (iCurX + iWordSize > iLineWidth)
		{
			iCurX = 0;
			iNumLines++;

			if (isspace(sString[i]))
				i++;

			if (i >= sString.length())
				break;

			if (iNumLines >= NumLines)
				return sString.c_str() + i;
		}

		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			int CharNum = sString[i] - ' ';

			iCurX += m_Characters[CharNum]->GetWidth()-4;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurX += 4*(m_Characters[0]->GetWidth()-4);
				} break;
			case '\n':
				{
					iCurX = 0;
					iNumLines++;

					if (iNumLines >= NumLines)
						return sString.c_str() + i + 1;

				} break;
			default: 
				break;
			}
		}
	}

	return "";
}

void Font2D::DrawStringDown(const std::string & sString, int x, int y,
						    unsigned int modColor)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in DrawStringDown()");

	int iCurX = 0;
	int iCurY = 0;

	int iTextHeight = m_Characters[0]->GetHeight();

	for (unsigned int i = 0; i < sString.length(); i++)
	{
		int CharNum = sString[i] - ' ';

		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			m_Characters[CharNum]->Draw(x+iCurX,y+iCurY-2,
				                        0,0,Image2D::Rot90CW,modColor);

			iCurY += m_Characters[CharNum]->GetWidth()-4;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurY += 4*(m_Characters[0]->GetWidth()-4);
				} break;
			case '\n':
				{
					iCurY = 0;
					iCurX -= iTextHeight;
				} break;
			default: 
				break;
			}
		}
	}
}

void Font2D::DrawStringUp(const std::string & sString, int x, int y,
						    unsigned int modColor)
{
	if (m_Characters.empty())
		throw IsisException("not initialized yet in DrawStringDown()");

	int iCurX = 0;
	int iCurY = 0;

	int iTextHeight = m_Characters[0]->GetHeight();

	for (int i = (int)sString.length()-1; i >= 0; i--)
	{
		int CharNum = sString[i] - ' ';

		if (sString[i] >= ' ' && sString[i] <= '~')
		{
			m_Characters[CharNum]->Draw(x+iCurX,y+iCurY-2,0,0,
	(Image2D::Flags)(Image2D::MirrorV|Image2D::MirrorH|Image2D::Rot90CW),
				modColor);

			iCurY += m_Characters[CharNum]->GetWidth()-4;
		}
		else
		{
			switch (sString[i])
			{
			case '\t':
				{
					iCurY += 4*(m_Characters[0]->GetWidth()-4);
				} break;
			case '\n':
				{
					iCurY = 0;
					iCurX -= iTextHeight;
				} break;
			default: 
				break;
			}
		}
	}
}

} // namespace Isis
