
#include <stdio.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/Font2D.h>

#include <isis/game/GUIGame.h>
#include <isis/game/GUIWindow.h>

#include <isis/game/IMouseHandler.h>
#include <isis/game/IKeyHandler.h>

namespace Isis {

#ifdef _DEBUG
int GUIGame::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,GUIGame::FuncType> GUIGame::m_ExportTable2;
std::map<std::string,GUIGame::ObjectType> GUIGame::m_ObjectTable;

GUIGame::GUIGame()
{
	if (m_ExportTable2.empty())
		FillExportTable2();

	if (m_ObjectTable.empty())
		FillObjectTable();

	m_Mouse.iX = -1;
	m_Mouse.iY = -1;
	m_Mouse.iOldX = -1;
	m_Mouse.iOldY = -1;

	m_pKeyboardFocusWindow = NULL;
	m_pMouseCaptureWindow = NULL;

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

GUIGame::~GUIGame()
{
#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int GUIGame::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void GUIGame::GameInit()
{
	GUIInit();
}

bool GUIGame::GameMain()
{
	m_pGraphicsEngine->Clear();

	std::list<GUIWindow*> orig_list;
	std::list<GUIWindow*>::iterator iter;

	for (iter = m_FrameUpdaters.begin();
		iter != m_FrameUpdaters.end();
		iter++)
	{
		orig_list.push_back(*iter);
		(*iter)->AddRef();
	}

	for (iter = orig_list.begin();
		iter != orig_list.end();
		iter++)
	{
		(*iter)->FrameUpdate();
		(*iter)->Release();
	}
	
	orig_list.clear();

	for (std::list<GUIWindow*>::iterator iter = m_WindowList.begin();
		iter != m_WindowList.end();
		iter++)
	{
		(*iter)->Draw();
	}

	return GUIMain();
}

void GUIGame::GameShutdown()
{
	GUIShutdown();

	SetKeyboardFocus(NULL);
	SetMouseCapture(NULL);

	while (!m_WindowList.empty())
	{
		GUIWindow * pWindow = m_WindowList.front();
		pWindow->Destroy();
	}

	while (GetDefaultFont())
		PopFont();
}

void GUIGame::KeyPressed(int key, ModifierKeys modifiers)
{
	GUIWindow * pWindow = GetKeyboardFocus();

	if (pWindow && pWindow->GetKeyHandler())
	{
		pWindow->GetKeyHandler()->HandleKey(key,modifiers,1);
	}
}

void GUIGame::KeyReleased(int key, ModifierKeys modifiers)
{
	GUIWindow * pWindow = GetKeyboardFocus();

	if (pWindow && pWindow->GetKeyHandler())
	{
		pWindow->GetKeyHandler()->HandleKey(key,modifiers,0);
	}
}

void GUIGame::MouseMoved(int x, int y)
{
	m_Mouse.iOldX = m_Mouse.iX;
	m_Mouse.iOldY = m_Mouse.iY;

	m_Mouse.iX = x;
	m_Mouse.iY = y;

	GUIWindow * pWindow = GetMouseCapture();

	if (!pWindow)
		pWindow = WindowFromPoint(x,y);

	if (!pWindow)
		return;

	pWindow->ScreenToWindow(x,y);

	if (pWindow && pWindow->GetMouseHandler())
	{
		pWindow->GetMouseHandler()->HandleMouse(x,y,GameApp::NonClick);
	}
}

void GUIGame::MouseClicked(ClickType click, int x, int y)
{
	m_Mouse.iOldX = m_Mouse.iX;
	m_Mouse.iOldY = m_Mouse.iY;

	m_Mouse.iX = x;
	m_Mouse.iY = y;

	GUIWindow * pWindow = GetMouseCapture();

	if (!pWindow)
		pWindow = WindowFromPoint(x,y);

	if (!pWindow)
	{
		EmptyClick(click,x,y);
		return;
	}

	pWindow->ScreenToWindow(x,y);

	if (pWindow && pWindow->GetMouseHandler())
	{
		pWindow->GetMouseHandler()->HandleMouse(x,y,click);
	}
}

void GUIGame::EmptyClick(ClickType click, int x, int y)
{
}

GUIWindow * GUIGame::WindowFromPoint(int x, int y, bool bCountTransparent)
{
	GUIWindow * pWindow = GetExclusiveWindow();

	if (pWindow)
	{
		return pWindow->WindowFromPoint(x-pWindow->GetX(),
										y-pWindow->GetY(),
										bCountTransparent);
	}

	std::list<GUIWindow*>::reverse_iterator riter;

	for (riter = m_WindowList.rbegin();
		 riter != m_WindowList.rend();
		 riter++)
	{
		int newx = x - (*riter)->GetX();
		int newy = y - (*riter)->GetY();

		pWindow = (*riter)->WindowFromPoint(newx,newy,bCountTransparent);

		if (pWindow)
			break;
	}

	return pWindow;
}

GUIWindow * GUIGame::MakeWindow(int x, int y, int width, int height)
{
	GUIWindow * pWindow = new GUIWindow(this,this);
	pWindow->AddRef();
	m_WindowList.push_back(pWindow);

	pWindow->Init(x,y,width,height);

	return pWindow;
}

void GUIGame::RemoveWindow(GUIWindow * pWindow)
{
	pWindow->Release();

	m_WindowList.remove(pWindow);
}

void GUIGame::DestroyAllWindows()
{
	while (m_WindowList.size() > 0)
		m_WindowList.front()->Destroy();
}

void GUIGame::WindowToScreen(int & x, int & y)
{
	// do nothing
}

void GUIGame::ScreenToWindow(int & x, int & y)
{
	// do nothing
}

void GUIGame::AddExclusiveWindow(GUIWindow * pWindow)
{
	m_ExclusiveStack.push_back(pWindow);
}

void GUIGame::RemoveExclusiveWindow(GUIWindow * pWindow)
{
	m_ExclusiveStack.remove(pWindow);
}

GUIWindow * GUIGame::GetExclusiveWindow()
{
	if (m_ExclusiveStack.empty())
		return NULL;

	return m_ExclusiveStack.back();
}

void GUIGame::AddFrameUpdater(GUIWindow * pWindow)
{
	m_FrameUpdaters.push_back(pWindow);
}

void GUIGame::RemoveFrameUpdater(GUIWindow * pWindow)
{
	m_FrameUpdaters.remove(pWindow);
}

bool GUIGame::UnhandledException(IsisException & ex)
{
	return false;
}

} // namespace Isis
