
#include <isis/game/GraphicsEngine.h>
#include <isis/game/Image2D.h>
#include <isis/game/Font2D.h>

#include <isis/game/GUIGame.h>
#include <isis/game/GUIWindow.h>

#include <isis/game/IWindowContainer.h>
#include <isis/game/IMouseHandler.h>
#include <isis/game/IKeyHandler.h>
#include <isis/game/ICustomDraw.h>
#include <isis/game/IDestroyCallback.h>
#include <isis/game/IFrameUpdate.h>

namespace Isis {

#ifdef _DEBUG
int GUIWindow::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,GUIWindow::FuncType> GUIWindow::m_ExportTable;

GUIWindow::GUIWindow(GUIGame * pGUIGame, IWindowContainer * pContainer)
{
	if (m_ExportTable.empty())
		FillExportTable();

	m_pGUIGame = pGUIGame;
	m_pGUIGame->AddRef();

	m_pContainer = pContainer;
	m_pContainer->AddRef();

	m_iX = 0;
	m_iY = 0;
	m_iWidth = 0;
	m_iHeight = 0;

	m_bExclusive = false;
	m_bTransparent = false;
	m_bHidden = false;

	m_Color = 0;

	m_pImage = NULL;
	m_ImageFlags = Image2D::None;

	m_TextColor = 0xFF000000;
	m_pFont = NULL;
	m_iTextDirection = 0;

	m_pMouseHandler = NULL;
	m_pKeyHandler = NULL;
	m_pCustomDraw = NULL;
	m_pDestroyCallback = NULL;
	m_pFrameUpdate = NULL;

	m_bDestroyed = false;

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

GUIWindow::~GUIWindow()
{
	if (m_pMouseHandler ||
		m_pKeyHandler ||
		m_pCustomDraw ||
		m_pDestroyCallback ||
		m_pFrameUpdate ||
		m_pImage ||
		m_pFont || 
		m_pGUIGame || 
		m_pContainer)
	{
		throw IsisException("GUIWindow object not destroyed properly");
	}

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int GUIWindow::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void GUIWindow::Init(int x, int y, int width, int height)
{
	m_iX = x;
	m_iY = y;

	m_iWidth = width;
	m_iHeight = height;

	m_pFont = m_pGUIGame->GetDefaultFont();

	if (m_pFont)
		m_pFont->AddRef();
}

void GUIWindow::Destroy()
{
	if (m_bDestroyed)
		return;

    if (m_pDestroyCallback)
    {
        m_pDestroyCallback->DestroyCallback();
		m_pDestroyCallback->Release();
        m_pDestroyCallback = NULL;
    }

    if (m_pFrameUpdate)
    {
		m_pFrameUpdate->Release();
		m_pGUIGame->RemoveFrameUpdater(this);
        m_pFrameUpdate = NULL;
    }

	m_bDestroyed = true;

	while (!m_WindowList.empty())
	{
		GUIWindow * pWindow = m_WindowList.front();
		pWindow->Destroy();
	}

	if (m_bExclusive)
	{
		m_pGUIGame->RemoveExclusiveWindow(this);
		m_bExclusive = false;
	}

	if (m_pMouseHandler)
	{
		m_pMouseHandler->Release();
		m_pMouseHandler = NULL;
	}

	if (m_pKeyHandler)
	{
		m_pKeyHandler->Release();
		m_pKeyHandler = NULL;
	}

	if (m_pCustomDraw)
	{
		m_pCustomDraw->Release();
		m_pCustomDraw = NULL;
	}

	if (m_pImage)
	{
		m_pImage->Release();
		m_pImage = NULL;
	}

	if (m_pFont)
	{
		m_pFont->Release();
		m_pFont = NULL;
	}

	if (m_pGUIGame)
	{
		if (this == m_pGUIGame->GetKeyboardFocus())
			m_pGUIGame->SetKeyboardFocus(NULL);

		if (this == m_pGUIGame->GetMouseCapture())
			m_pGUIGame->SetMouseCapture(NULL);

		m_pGUIGame->Release();
		m_pGUIGame = NULL;
	}

	if (m_pContainer)
	{
		IWindowContainer * pContainer = m_pContainer;
		m_pContainer = NULL;
		pContainer->RemoveWindow(this);
		pContainer->Release();
	}
}

GUIWindow * GUIWindow::MakeWindow(int x, int y, int width, int height)
{
	GUIWindow * pWindow = new GUIWindow(m_pGUIGame,this);
	pWindow->AddRef();

	pWindow->Init(x,y,width,height);

	m_WindowList.push_back(pWindow);

	return pWindow;
}

void GUIWindow::RemoveWindow(GUIWindow * pWindow)
{
	pWindow->Release();

	m_WindowList.remove(pWindow);
}

void GUIWindow::WindowToScreen(int & x, int & y)
{
	x += m_iX;
	y += m_iY;

	if (m_pContainer)
		m_pContainer->WindowToScreen(x,y);
}

void GUIWindow::ScreenToWindow(int & x, int & y)
{
	x -= m_iX;
	y -= m_iY;

	if (m_pContainer)
		m_pContainer->ScreenToWindow(x,y);
}

void GUIWindow::Draw()
{
    if (IsHidden())
        return;

	int x = 0;
	int y = 0;

	WindowToScreen(x,y);

	GraphicsEngine * pGraphicsEngine = m_pGUIGame->GetGraphicsEngine();

	if (m_pImage)
		m_pImage->Draw(x,y,GetWidth(),GetHeight(),m_ImageFlags);

	if (m_Color & 0xFF000000)
		pGraphicsEngine->FillRectangle(x,y,m_iWidth,m_iHeight,m_Color);

	if (m_pCustomDraw)
		m_pCustomDraw->Draw(x,y,m_iWidth,m_iHeight,pGraphicsEngine);

	if (m_sText.length() && m_pFont)
	{
		if (m_iTextDirection == 0)
		{
			m_pFont->DrawString(m_sText,
								x,y,
								m_iWidth,m_iHeight,
								m_TextColor);
		}
		else if (m_iTextDirection == 1)
		{
			m_pFont->DrawStringDown(m_sText,
								    x,y,
								    m_TextColor);
		}
		else if (m_iTextDirection == 2)
		{
			m_pFont->DrawStringUp(m_sText,
								  x,y,
								  m_TextColor);
		}
	}

	for (std::list<GUIWindow *>::iterator iter = m_WindowList.begin();
		 iter != m_WindowList.end();
		 iter++)
	{
		(*iter)->Draw();
	}
}

GUIWindow * GUIWindow::WindowFromPoint(int x, int y, bool bCountTransparent)
{
	GUIWindow * pWindow = NULL;
	std::list<GUIWindow*>::reverse_iterator riter;

	for (riter = m_WindowList.rbegin();
		 riter != m_WindowList.rend();
		 riter++)
	{
		int newx = x - (*riter)->GetX();
		int newy = y - (*riter)->GetY();

		pWindow = (*riter)->WindowFromPoint(newx,newy,bCountTransparent);

		if (pWindow)
			break;
	}

	if (!pWindow)
	{
		if ((bCountTransparent || !IsTransparent()) && !IsHidden() && 
			x >= 0 && x < m_iWidth &&
			y >= 0 && y < m_iHeight)
		{
			pWindow = this;
		}
	}

	return pWindow;
}

void GUIWindow::FrameUpdate()
{
    if (m_pFrameUpdate)
        m_pFrameUpdate->FrameUpdate();
}

}  // namespace Isis
