
#include <SDL/SDL.h>

#include <isis/game/GameApp.h>
#include <isis/game/GraphicsEngine.h>
#include <isis/game/SoundEngine.h>

namespace Isis {

std::map<std::string,GameApp::FuncType> GameApp::m_ExportTable;

GameApp::GameApp()
{
	m_pGraphicsEngine = new GraphicsEngine();
	m_pGraphicsEngine->AddRef();
	
	m_pSoundEngine = new SoundEngine();
	m_pSoundEngine->AddRef();
	
    m_iFrameRate = 60;
    m_bUseFrameRate = true;
	m_sWindowTitle = "Isis Application";

	m_CurModifiers = NoKey;
    
    if (m_ExportTable.empty())
        FillExportTable();
}

GameApp::~GameApp()
{
    if (NULL != m_pGraphicsEngine)
    {
	    m_pGraphicsEngine->Destroy();
	    m_pGraphicsEngine->Release();
    }
	if (NULL != m_pSoundEngine)
    {
	    m_pSoundEngine->Destroy();
	    m_pSoundEngine->Release();
    }

    SDL_Quit();
}

void GameApp::Create()
{
    int result;

    result = SDL_Init(SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE);

    if (-1 == result)
        throw IsisException("Error initializing SDL");

    m_pGraphicsEngine->CreateDisplay();
	m_pSoundEngine->InitAudio();
    
	SetWindowTitle(m_sWindowTitle);

    GameInit();
}

void GameApp::SetWindowTitle(const std::string & sWindowTitle)
{
    m_sWindowTitle = sWindowTitle;

	SDL_WM_SetCaption(m_sWindowTitle.c_str(),NULL);
}

void GameApp::Run()
{
    SDL_Event event;
    bool done = false;
    
    int iOldTime = SDL_GetTicks();

    while (!done)
    {
		try
		{
        while (SDL_PollEvent(&event))
        {
			switch (event.type)
			{
			case SDL_KEYDOWN:
			case SDL_KEYUP:
				{
					int key = event.key.keysym.sym;

#ifdef _DEBUG
					if (key == SDLK_ESCAPE)
					{
						done = true;
						break;
					}
#endif 
    				if (key == SDLK_RETURN &&
                        event.type == SDL_KEYDOWN && 
							 (event.key.keysym.mod & KMOD_LALT || 
							  event.key.keysym.mod & KMOD_RALT))
					{
						m_pGraphicsEngine->ToggleFullscreen();
					}

					if (key == SDLK_LSHIFT ||
						key == SDLK_RSHIFT)
						m_CurModifiers = (ModifierKeys)
							(event.type == SDL_KEYDOWN ?
								m_CurModifiers|Shift : 
								m_CurModifiers&(~Shift));

					if (key == SDLK_LCTRL ||
						key == SDLK_RCTRL)
						m_CurModifiers = (ModifierKeys)
							(event.type == SDL_KEYDOWN ?
								m_CurModifiers|Control : 
								m_CurModifiers&(~Control));

					if (key == SDLK_LALT ||
						key == SDLK_RALT)
						m_CurModifiers = (ModifierKeys)
							(event.type == SDL_KEYDOWN ?
								m_CurModifiers|Alt : 
								m_CurModifiers&(~Alt));

					if (key == SDLK_LMETA ||
						key == SDLK_RMETA)
						m_CurModifiers = (ModifierKeys)
							(event.type == SDL_KEYDOWN ?
								m_CurModifiers|Meta : 
								m_CurModifiers&(~Meta));

					if (SDL_KEYDOWN == event.type)
						KeyPressed(key,m_CurModifiers);
					else
						KeyReleased(key,m_CurModifiers);

				} break;
			case SDL_MOUSEMOTION:
				{
					int x = event.motion.x;
					int y = event.motion.y;

					MouseMoved(x,y);

				} break;
			case SDL_MOUSEBUTTONDOWN:
				{
					ClickType click = NonClick;

					if (event.button.button == SDL_BUTTON_LEFT)
						click = LeftDown;
					else if (event.button.button == SDL_BUTTON_MIDDLE)
						click = MiddleDown;
					else
						click = RightDown;

					int x = event.button.x;
					int y = event.button.y;

					if (click == LeftDown && CheckModifier(Control))
						click = RightDown;

					MouseClicked(click,x,y);

				} break;
			case SDL_MOUSEBUTTONUP:
				{
					ClickType click = NonClick;

					if (event.button.button == SDL_BUTTON_LEFT)
						click = LeftUp;
					else if (event.button.button == SDL_BUTTON_MIDDLE)
						click = MiddleUp;
					else
						click = RightUp;

					int x = event.button.x;
					int y = event.button.y;

					if (click == LeftUp && CheckModifier(Control))
						click = RightUp;

					MouseClicked(click,x,y);

				} break;
			case SDL_QUIT:
				{
					done = true;
				} break;
			default:
				break;
			}
        }

        if (done)
            break;

        if (!GameMain())
			break;

        m_pGraphicsEngine->Present();

        int iNewTime = SDL_GetTicks();

        while ((float)(iNewTime - iOldTime) < (1000.0f / (float)m_iFrameRate))
        {
            SDL_Delay(0);
            iNewTime = SDL_GetTicks();
        }

        iOldTime = iNewTime;

		}
		// if an IsisException occurs, send it to the unhandled handler
		catch (Isis::IsisException & ie)
		{
			done = !UnhandledException(ie);
		}
    }

    GameShutdown();
}

bool GameApp::CheckModifier(ModifierKeys key)
{
	return (NoKey != (m_CurModifiers & key));
}

void GameApp::KeyPressed(int key, ModifierKeys modifiers)
{
}

void GameApp::KeyReleased(int key, ModifierKeys modifiers)
{
}

void GameApp::MouseMoved(int x, int y)
{
}

void GameApp::MouseClicked(ClickType click, int x, int y)
{
}

bool GameApp::UnhandledException(IsisException & ex)
{
	return false;
}

} // namespace Isis
