
#ifdef WIN32
#include <windows.h>
#endif // WIN32

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/VRAMObject.h>

#include <isis/game/Image2D.h>
#include <isis/game/Texture.h>
#include <isis/game/Font2D.h>

#include <isis/game/VectorAnimation.h>

namespace Isis {

#ifdef _DEBUG
int GraphicsEngine::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,GraphicsEngine::FuncType> GraphicsEngine::m_ExportTable;
std::map<std::string,GraphicsEngine::ObjectType> 
										GraphicsEngine::m_ObjectTable;

GraphicsEngine::GraphicsEngine()
{
	m_bDestroyed = false;

    if (m_ExportTable.empty())
        FillExportTable();

	if (m_ObjectTable.empty())
		FillObjectTable();

    m_iWidth = 1024;
    m_iHeight = 768;
    m_bWindowed = true;

    m_bTexMapEnabled = false;
    m_bOrtho = false;
	m_bDepth = false;
	m_bBlend = false;

    m_pBoundTexture = NULL;
    
    m_pMainSurface = NULL;

	m_Color = 0;

	AddImagePrefix("");

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

GraphicsEngine::~GraphicsEngine()
{
	if (!m_bDestroyed)
		throw IsisException("GraphicsEngine not destroyed before ~");

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int GraphicsEngine::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void GraphicsEngine::Destroy()
{
	m_bDestroyed = true;

    if (m_pBoundTexture)
        m_pBoundTexture->Release();
}

void GraphicsEngine::CreateDisplay()
{
    int result;
	bool bWasInit = false;

    if (!SDL_WasInit(SDL_INIT_VIDEO))
    {
        result = SDL_InitSubSystem(SDL_INIT_VIDEO);

        if (result < 0)
            throw IsisException("Erorr initializing SDL video subsytem:",
                                SDL_GetError());
    }
	else
		bWasInit = true;

 	SDL_GL_SetAttribute(SDL_GL_RED_SIZE,   5);
    SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
    SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,  5);
    SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE, 16);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16 );
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    
    if (m_bWindowed)
	{
#ifdef WIN32
		if (GetSystemMetrics(SM_CXSCREEN) == m_iWidth && 
			GetSystemMetrics(SM_CYSCREEN) == m_iHeight && 
			!bWasInit)
		{
			m_bWindowed = false;
		}
		else
		{
			m_pMainSurface = SDL_SetVideoMode(m_iWidth,m_iHeight,0,SDL_OPENGL);
		}
#else
        m_pMainSurface = SDL_SetVideoMode(m_iWidth,m_iHeight,0,SDL_OPENGL);
#endif
	}

	if (!m_bWindowed)
        m_pMainSurface = SDL_SetVideoMode(m_iWidth,
							m_iHeight,32,SDL_OPENGL|SDL_FULLSCREEN);

    if (NULL == m_pMainSurface)
        throw IsisException("Erorr calling SDL_SetVideoMode");
    
    glViewport(0,0,m_iWidth,m_iHeight);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	SetColor(0xFFFFFFFF);
}

void GraphicsEngine::ToggleFullscreen()
{
	m_bWindowed = !m_bWindowed;

	SetOrtho(false);
	EnableTexMap(false);
	SetBlend(false);
	SetDepth(false);

    BindTexture(NULL);

    std::list<VRAMObject *>::iterator iter;
    
    for (iter = m_VRAMList.begin();
         iter != m_VRAMList.end();
         iter++)
    {
        (*iter)->UnloadVRAM();
    }

    CreateDisplay();

    for (iter = m_VRAMList.begin();
         iter != m_VRAMList.end();
         iter++)
    {
        (*iter)->LoadVRAM();
    }

	RebindCurrentTexture();
}

void GraphicsEngine::Present()
{
    SDL_GL_SwapBuffers();
}

void GraphicsEngine::Clear(unsigned int uiColor)
{
   glClearColor((float)((uiColor&0x00FF0000)>>16)/255.0f,
                (float)((uiColor&0x0000FF00)>>8)/255.0f,
                (float)(uiColor&0x000000FF)/255.0f,
                (float)((uiColor&0xFF000000)>>24)/255.0f);

   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
}

Texture * GraphicsEngine::LoadIGZ(const std::string & sImageFilename,
								  std::vector<Image2D*> & SubImages)
{
    Texture * pTexture = new Texture(this);
    
    pTexture->Init(sImageFilename,SubImages);
    
    return pTexture;
}

Texture * GraphicsEngine::LoadIGZ(const std::string & sImageFilename)
{
	std::vector<Image2D*> SubImages;
    Texture * pTexture = new Texture(this);
    
	pTexture->AddRef();
    pTexture->Init(sImageFilename,SubImages);

	for (int i = 0; i < (int)SubImages.size(); i++)
	{
		SubImages[i]->AddRef();
		SubImages[i]->Release();
	}

	pTexture->SafeRelease();
    return pTexture;
}

void GraphicsEngine::LoadVGZ(const std::string & sImageFilename,
							 std::vector<VectorAnimation*> & animations)
{
	VectorAnimation::LoadVGZ(this,sImageFilename,animations);
}

void GraphicsEngine::BindTexture(Texture * pTexture)
{
	if (m_pBoundTexture != pTexture)
	{
		if (m_pBoundTexture)
			m_pBoundTexture->Release();
	    
		m_pBoundTexture = pTexture;
	    
		if (m_pBoundTexture)
		{
			m_pBoundTexture->AddRef();
	    
			glBindTexture(GL_TEXTURE_2D,m_pBoundTexture->GetID());
		}
		else
			glBindTexture(GL_TEXTURE_2D,0);
	}
}

void GraphicsEngine::RebindCurrentTexture()
{
    if (m_pBoundTexture)
        glBindTexture(GL_TEXTURE_2D,m_pBoundTexture->GetID());
    else
        glBindTexture(GL_TEXTURE_2D,0);
}

Font2D * GraphicsEngine::LoadFont(const std::string & sIGZFilename)
{
	Font2D * pFont = new Font2D(this);
	pFont->Init(sIGZFilename);
    
    return pFont;
}

unsigned int GraphicsEngine::Color(unsigned char red,
								   unsigned char green,
								   unsigned char blue,
								   unsigned char alpha)
{
	return (alpha<<24) | (red<<16) | (green<<8) | blue;
}

void GraphicsEngine::DrawRectangle(int x, int y, 
								   int width, int height, 
								   unsigned int color)
{
	SetColor(color);

    EnableTexMap(false);
    SetOrtho(true);
    SetBlend(true);
	SetDepth(false);

	int x1 = x;
	int y1 = y;
	int x2 = x1 + width;
	int y2 = y1 + height;

	glBegin(GL_LINE_STRIP);

	glVertex2i(x1,y1);
	glVertex2i(x2,y1);
	glVertex2i(x2,y2);
	glVertex2i(x1,y2);
	glVertex2i(x1,y1);

	glEnd();
}

void GraphicsEngine::FillRectangle(int x, int y,
								   int width, int height,
								   unsigned int color)
{
	SetColor(color);

    EnableTexMap(false);
    SetOrtho(true);
    SetBlend(true);
	SetDepth(false);

	int x1 = x;
	int y1 = y;
	int x2 = x1 + width;
	int y2 = y1 + height;

	glBegin(GL_TRIANGLE_STRIP);

	glVertex2i(x1,y1);
	glVertex2i(x1,y2);
	glVertex2i(x2,y1);
	glVertex2i(x2,y2);

	glEnd();
}

void GraphicsEngine::DrawLine(int x1, int y1,
							  int x2, int y2, 
							  unsigned int color)
{
	SetColor(color);

    EnableTexMap(false);
    SetOrtho(true);
    SetBlend(true);
	SetDepth(false);

	glBegin(GL_LINES);

	glVertex2i(x1,y1);
	glVertex2i(x2,y2);

	glEnd();
}

void GraphicsEngine::SetColor(unsigned int color)
{
	if (color != m_Color)
	{
		m_Color = color; 

		glColor4ub((color&0x00FF0000)>>16,
					(color&0x0000FF00)>>8,
					(color&0x000000FF)>>0,
					(color&0xFF000000)>>24);
	}
}

} // namespace Isis
