
#include <algorithm>
#include <math.h>

#include <SDL/SDL_opengl.h>

#include <isis/game/ParticleManager.h>
#include <isis/game/ParticleSource.h>

#define RAND_RANGE_I(a,b)		((a)==(b)?(a):(a)+rand()%((b)-(a)))
#define RAND_RANGE_F(a,b)		((a)==(b)?(a):(a)+((float)rand()/RAND_MAX)*((b)-(a)))

namespace Isis {

#ifdef _DEBUG
int ParticleManager::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,ParticleManager::FuncType> ParticleManager::m_ExportTable;

ParticleManager::ParticleManager()
{
    if (m_ExportTable.empty())
        FillExportTable();

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG

	m_uiListID = glGenLists(1);
	glNewList(m_uiListID,GL_COMPILE);

	glBegin(GL_TRIANGLE_FAN);

	glVertex3f(0,0,0);

	for (float a = 360.0f; a > 360.0f/(2*8.0f); a -= 360.0f/8.0f)
	{
		glVertex3d(cos(a*3.141592/180.0),1*sin(a*3.141592/180.0),0);
	}
	glVertex3f(1,0,0);

	glEnd();

	glEndList();
}

ParticleManager::~ParticleManager()
{
	for (std::set<Particle *>::iterator i = m_Particles.begin();
		 i != m_Particles.end();
		 i++)
	{
		(*i)->source->Release();
		delete *i;
	}

	for (std::set<ParticleSource *>::iterator i = m_Sources.begin();
		 i != m_Sources.end();
		 i++)
	{
		(*i)->Release();
	}

	glDeleteLists(m_uiListID,1);

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int ParticleManager::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void ParticleManager::Init()
{
}

void ParticleManager::AddSource(ParticleSource * pSource)
{
	pSource->AddRef();
	m_Sources.insert(pSource);
}

void ParticleManager::Draw(Isis::GraphicsEngine * pGraphicsEngine)
{
	pGraphicsEngine->SetBlend(true);
	pGraphicsEngine->SetDepth(false);
//	pGraphicsEngine->SetOrtho(true);
	pGraphicsEngine->EnableTexMap(false);

	pGraphicsEngine->SetColor(0);
	
	for (std::set<Particle *>::iterator i = m_Particles.begin();
		 i != m_Particles.end();
		 i++)
	{
		glPushMatrix();

		Particle * p = *i;

		pGraphicsEngine->SetColor(p->source->CalcColor(
			p->counter%p->lifetime));

		glTranslatef(p->x,p->y,0);
		glScalef(p->size,p->size,1);

		glCallList(m_uiListID);

		glPopMatrix();
	}
}

void ParticleManager::Update()
{
	for (std::set<ParticleSource *>::iterator i = m_Sources.begin();
		 i != m_Sources.end();  /* nop */ )
	{
		ParticleSource::Desc & d = (*i)->GetDesc();

		d.x += d.xv;
		d.y += d.yv;

		d.xv += d.xa;
		d.yv += d.ya;

		if (d.active)
		{
			if (d.p_prob >= 1.0f)
			{
				if (rand()%(int)d.p_prob == 0)
                    NewParticle(*i);
			}
			else
			{
				for (float f = 1.0f/d.p_prob; f >= 1.0f; f -= 1.0f)
					NewParticle(*i);

				if (rand()%100 >= (int)(d.p_prob*100))
					NewParticle(*i);
			}
		}

		if (d.active && d.expire_time > 0)
			d.expire_time--;

		if (d.expire_time == 0)
		{
			(*i)->Release();
			m_Sources.erase(i++);
		}
		else
			++i;
	}

	for (std::set<Particle *>::iterator i = m_Particles.begin();
		 i != m_Particles.end(); /* nop */ )
	{
		Particle * p = *i;

		p->x += p->xv;
		p->y += p->yv;

		p->xv += p->xa;
		p->yv += p->ya;

		p->counter++;

		if (p->counter >= p->lifetime)
		{
			p->source->Release();
			delete *i;
			m_Particles.erase(i++);
		}
		else
			++i;
	}
}

void ParticleManager::NewParticle(ParticleSource * source)
{
	Particle * p = new Particle();
	p->source = source;
	p->source->AddRef();

	ParticleSource::Desc d = source->GetDesc();

	p->lifetime = RAND_RANGE_I(d.p_life_min,d.p_life_max);
	p->counter = 0;

	p->size = RAND_RANGE_F(d.p_size_min,d.p_size_max);

	p->x = d.x;
	p->y = d.y;

	p->xv = d.xv + RAND_RANGE_F(d.p_xv_min,d.p_xv_max);
	p->yv = d.yv + RAND_RANGE_F(d.p_yv_min,d.p_yv_max);

	p->xa = d.p_xa;
	p->ya = d.p_ya;

	m_Particles.insert(p);
}

} // namespace Isis
