
#ifdef WIN32
#include <windows.h>
#endif // WIN32

#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include <isis/game/SoundEngine.h>
#include <isis/game/Sound.h>
//SDL Audio


namespace Isis {
//It's not a global..it's a transcontinental...
void *m_pMusic;

#ifdef _DEBUG
int SoundEngine::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,SoundEngine::FuncType> SoundEngine::m_ExportTable;
std::map<std::string,SoundEngine::ObjectType> 
										SoundEngine::m_ObjectTable;

SoundEngine::SoundEngine()
{
	m_bDestroyed = false;

    if (m_ExportTable.empty())
        FillExportTable();


	// What is an object table?
	if (m_ObjectTable.empty())
		FillObjectTable();


    m_iAudioRate = 11025;        //okay rate...
	m_AudioFormat = AUDIO_S16; //16 bit audio 
	m_iAudioChannels = 1; // Mono (note *not* mixing channels)
	m_iAudioBuffers = 4096; //4k buffer
	m_iMixingChannels=16; //Start with one channel for sound effects
    m_iSoundVolume=MIX_MAX_VOLUME/2;
	m_iMusicVolume=MIX_MAX_VOLUME/2;
	Mix_AllocateChannels(m_iMixingChannels);
	Mix_Volume(-1,m_iSoundVolume);
	Mix_VolumeMusic(m_iMusicVolume);

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

SoundEngine::~SoundEngine()
{
	if (!m_bDestroyed)
		throw IsisException("SoundEngine not destroyed before ~");

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int SoundEngine::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void SoundEngine::Destroy()
{
	m_bDestroyed = true;

	//I think that so
	//Release Music and Sounds!

    //if (m_pBoundTexture)
      //  m_pBoundTexture->Release();
}

void SoundEngine::InitAudio()
{
    int result;
	bool bWasInit = false;

    if (!SDL_WasInit(SDL_INIT_AUDIO))
    {
        result = SDL_InitSubSystem(SDL_INIT_AUDIO);

        if (result < 0)
            throw IsisException("Erorr initializing SDL audio subsytem:",
                                SDL_GetError());
    }
	else
		bWasInit = true;
	if(Mix_OpenAudio
		(m_iAudioRate, m_AudioFormat, m_iAudioChannels, m_iAudioBuffers))
            throw IsisException("Erorr opening SDL audio device:",
                                SDL_GetError());

}


Sound * SoundEngine::LoadSound(const std::string & sWAVEFilename)
{
	Sound * pSound = new Sound(this);
	pSound->Init(sWAVEFilename);
	/*We'll want to make sure we don't reload loaded sounds in the future*/
	pSound->AddRef();
	
    return pSound;
}

//Set the sound effects volume to Volume. Note that a value
// of zero sets all sounds to zero. 
void SoundEngine::SetSoundVolume(int Volume)
{
	if(Volume>MIX_MAX_VOLUME)
	{
		Mix_Volume(-1,MIX_MAX_VOLUME);
		m_iSoundVolume=MIX_MAX_VOLUME;
	}
	else if(Volume<=0)
	{
		Mix_Volume(-1, 0);
		m_iSoundVolume=0;
	}
	else
	{
		Mix_Volume(-1,Volume);
		m_iSoundVolume=Volume;
	}
}
int SoundEngine::GetSoundVolume()
{
	return m_iSoundVolume;
}

void SoundEngine::SetMusicVolume(int Volume)
{
	if(Volume>MIX_MAX_VOLUME)
	{
		Mix_VolumeMusic(MIX_MAX_VOLUME);
	}
	else if(Volume<=0)
	{
		Mix_VolumeMusic(0);
		m_iMusicVolume=0;
	}
	else
	{
		Mix_VolumeMusic(Volume);
		m_iMusicVolume=Volume;
	}
}
int SoundEngine::GetMusicVolume()
{
	return m_iMusicVolume;
}

void SoundEngine::Play(Sound *pSound,int iRepeat)
{
	int channel=-1;
	if(!pSound)	
		throw IsisException("Error attempting to play NULL sound");
	channel=Mix_PlayChannel(-1,pSound->GetChunk(),iRepeat);
	if(channel<0)
	{
		m_iMixingChannels+=10;
		Mix_AllocateChannels(m_iMixingChannels);
		if((channel=Mix_PlayChannel(-1,pSound->GetChunk(),iRepeat))<0)
		{
#ifdef _DEBUG			
			printf("Error playing pSound: %s",pSound->GetName(),
			std::string(Mix_GetError()));
#endif
		}
	}
	Mix_Volume(channel,m_iSoundVolume);

	/* This was going to keep track of what's playing where.
	   Not a necessary feature yet
	std::list<int>::iterator iter;
	for (iter = m_ChannelsList.begin(); iter != m_ChannelsList.end(); iter++)
    {
     if(!Mix_Playing(*iter))
	 {
		 if(!Mix_Paused(*iter))
			
			//Free channel
	*/

}
void SoundEngine::FreeChunk(Mix_Chunk *p_Chunk)
{
	//we could wait until it's done playing
	//but instead we'll just print an error in debug mode
	for(int i=0;i<Mix_Playing(-1);i++)
        if(p_Chunk==Mix_GetChunk(i))
		{
		#ifdef _DEBUG
			printf("FreeChunk: killing channel early\n");
		#endif
			Mix_HaltChannel(i);
		}

	Mix_FreeChunk(p_Chunk);
}

void SoundEngine::PlayMusic(const std::string  & sWAVEFilename,
					void (*p_MusicFinished)())
{
	if(m_pMusic != NULL) {
	
		/* Stop the music from playing */
		Mix_HaltMusic();
		if(m_pMusic!=NULL)
		{
			Mix_FreeMusic((Mix_Music *)m_pMusic);
			m_pMusic=NULL;
		}
	}
	
	if(sWAVEFilename=="")return;
	/* load up the music */
	m_pMusic = Mix_LoadMUS(sWAVEFilename.c_str());
	if(m_pMusic==NULL)
	{
#ifdef _DEBUG
		printf("Error, music file failed to load %s",sWAVEFilename.c_str());
#endif
	return;
	}
	//Music loops
	Mix_PlayMusic((Mix_Music *)m_pMusic, -1);

	/* Add a hook to free the music */
	Mix_HookMusicFinished(m_MusicFinished);
}
	
//Run this function on completion of music NULL disables
void SoundEngine::HookMusicFinished(void (*p_MusicFinished)())
{
}
void SoundEngine::m_MusicFinished()
{
	if(m_pMusic!=NULL)
	{
		Mix_FreeMusic((Mix_Music *)m_pMusic);
		m_pMusic=NULL;
	}
}


} // namespace Isis
