
#include <math.h>
#include <sys/stat.h>
#include <zlib.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include <algorithm>

#include <isis/util.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/Texture.h>
#include <isis/game/CellImage.h>

namespace Isis {

#ifdef _DEBUG
int Texture::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,Texture::FuncType> Texture::m_ExportTable;

Texture::Texture(GraphicsEngine * pGraphicsEngine)
{
    if (m_ExportTable.empty())
        FillExportTable();

    m_uiID = 0;
    m_iWidth = 0;
	m_iHeight = 0;

	m_pGraphicsEngine = pGraphicsEngine;
	m_pGraphicsEngine->AddRef();

    m_pGraphicsEngine->VRAMObjectCreated(this);

    m_pData = NULL;

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

Texture::~Texture()
{
    if (NULL != m_pData)
    {
        delete [] m_pData;
        m_pData = NULL;
    }

	UnloadVRAM();

	if (m_pGraphicsEngine)
    {
        m_pGraphicsEngine->VRAMObjectDestroyed(this);
		m_pGraphicsEngine->Release();
    }

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int Texture::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

int Texture::ReadInt(gzFile File)
{
	unsigned char bytes[4];

	if (4 != gzread(File,bytes,4))
	{
		gzclose(File);
		throw IsisException("Error reading image file");
	}

	return bytes[0] | (bytes[1] << 8) | 
		   (bytes[2] << 16) | (bytes[3] << 24);
}

void Texture::Init(const std::string & sImageFilename,
				   std::vector<Image2D*> & SubImages)
{
    struct stat statbuf;
	std::string prefix;
	bool bFound = false;
	std::vector<std::string> Prefixes = m_pGraphicsEngine->GetImagePrefixes();

    if (NULL != m_pData)
        throw IsisException("Already initialized in Init");

	for (unsigned int i = 0; i < Prefixes.size(); i++)
	{
		prefix = Prefixes[i];

		if (-1 != stat((prefix + sImageFilename).c_str(),&statbuf))
		{
			bFound = true;
			break;
		}
	}

	if (!bFound)
		throw IsisException("Could not stat image file:",
							sImageFilename.c_str());

	gzFile file = gzopen((prefix + sImageFilename).c_str(),"rb");
	
	if (NULL == file)
		throw IsisException("Could not open image file");

	m_iWidth = ReadInt(file);
	m_iHeight = ReadInt(file);

	int iNumSubImages = ReadInt(file);

	if (iNumSubImages < 0 || iNumSubImages > 4096)
	{
		gzclose(file);
		throw IsisException("Error reading image file");
	}

	for (int i = 0; i < iNumSubImages; i++)
	{
		int iSIx = ReadInt(file);
		int iSIy = ReadInt(file);
		int iSIwidth = ReadInt(file);
		int iSIheight = ReadInt(file);

		ReadInt(file); // borders (unused here)

		CellImage * pCellImage = new CellImage();
		pCellImage->Init(this,iSIx,iSIy,iSIwidth,iSIheight,Image2D::None);
		SubImages.push_back(pCellImage);
	}

	int iDataSize = m_iWidth * m_iHeight * 4;

	if (iDataSize <= 0 || iDataSize > 500000000)
	{
		gzclose(file);
		throw IsisException("Image size is invalid");
	}
	
	m_pData = new unsigned char[iDataSize];
	
	if (iDataSize != (int)gzread(file,m_pData,iDataSize))
	{
		gzclose(file);
		delete [] m_pData;

		throw IsisException("Error reading image data");
	}

	gzclose(file);

	LoadVRAM();

	m_pGraphicsEngine->RebindCurrentTexture();
}

void Texture::Init(unsigned char * pData, int width, int height)
{
    if (NULL != m_pData)
        throw IsisException("Already initialized in Init");

	m_iWidth = width;
	m_iHeight = height;

	int iDataSize = m_iWidth * m_iHeight * 4;

	m_pData = new unsigned char [iDataSize];

	memcpy(m_pData,pData,iDataSize);

	LoadVRAM();

	m_pGraphicsEngine->RebindCurrentTexture();
}

void Texture::UnloadVRAM()
{
	if (m_uiID != 0)
    {
		glDeleteTextures(1,(GLuint*)&m_uiID);
        m_uiID = 0;
    }
}

void Texture::LoadVRAM()
{
    if (NULL == m_pData || m_uiID != 0)
        throw IsisException("Not initialized in RecreateTexture()");

	glGenTextures(1,(GLuint*)&m_uiID);
	glBindTexture(GL_TEXTURE_2D,m_uiID);

	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

	glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,m_iWidth,m_iHeight,0,
			     GL_RGBA,GL_UNSIGNED_BYTE,m_pData);
}

void Texture::Draw(int iX, int iY, 
				   int iWidth, int iHeight,
				   Image2D::Flags DrawFlags,
				   unsigned int modColor,
				   int iCellX, int iCellY,
				   int iCellWidth, int iCellHeight)
{
    m_pGraphicsEngine->SetOrtho(true);
	m_pGraphicsEngine->SetDepth(false);

	Draw3D(iX,iY,iWidth,iHeight,DrawFlags,modColor,
		   iCellX,iCellY,iCellWidth,iCellHeight);
}

void Texture::Draw3D(int iX, int iY, 
				     int iWidth, int iHeight,
				     Image2D::Flags DrawFlags,
				     unsigned int modColor,
				     int iCellX, int iCellY,
				     int iCellWidth, int iCellHeight)
{
	m_pGraphicsEngine->SetBlend(true);
    m_pGraphicsEngine->EnableTexMap(true);
	m_pGraphicsEngine->BindTexture(this);

	int iGEWidth = m_pGraphicsEngine->GetWidth();
	int iGEHeight = m_pGraphicsEngine->GetHeight();

	int iTexWidth = m_iWidth;
	int iTexHeight = m_iHeight;

	int iDrawWidth = m_iWidth;
	int iDrawHeight = m_iHeight;

	if (iCellWidth == 0 || iCellHeight == 0)
	{
		iCellWidth = iDrawWidth;
		iCellHeight = iDrawHeight;

		if ((DrawFlags & Image2D::TileExtra) && 
			(iWidth > iDrawWidth) && 
			(iHeight > iDrawHeight))
		{
			Draw(iX,iY,iWidth,iHeight,
				(Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
				modColor,0,0,iWidth,iHeight);
			return;
		}
	}
	else
	{
		iDrawWidth = iCellWidth;
		iDrawHeight = iCellHeight;
	}

	if (iWidth == 0 || iHeight == 0)
	{
		iWidth = iDrawWidth;
		iHeight = iDrawHeight;
	}

	if (iWidth != iDrawWidth || 
		iHeight != iDrawHeight)
	{
		if (DrawFlags & Image2D::TileExtra)
		{
			if (!(DrawFlags & Image2D::Rot90CW))
			{
				int iXOff, iYOff;

				for (iXOff = 0; 
					 iXOff < iWidth - iDrawWidth; 
					 iXOff += iDrawWidth)
				{
					for (iYOff = 0; 
						 iYOff < iHeight - iDrawHeight; 
						 iYOff += iDrawHeight)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX, iCellY,
							 iCellWidth, iCellHeight);
					}

					if (iHeight - iYOff > 0)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX,
							 iCellY + (DrawFlags & Image2D::MirrorV ? 
									  iCellHeight - iHeight + iYOff : 0),
							 iDrawWidth,
							 iHeight - iYOff);
					}
				}

				if (iWidth - iXOff > 0)
				{
					for (iYOff = 0; 
						 iYOff < iHeight - iDrawHeight; 
						 iYOff += iDrawHeight)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX + (DrawFlags & Image2D::MirrorH ? 
									  iCellWidth - iWidth + iXOff : 0),
							 iCellY,
							 iWidth - iXOff,
							 iCellHeight);
					}

					if (iHeight - iYOff > 0)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX + (DrawFlags & Image2D::MirrorH ? 
									  iCellWidth - iWidth + iXOff : 0),
							 iCellY + (DrawFlags & Image2D::MirrorV ? 
									  iCellHeight - iHeight + iYOff : 0),
							 iWidth - iXOff,
							 iHeight - iYOff);
					}
				}
			}
			else
			{
				int iXOff, iYOff;

				for (iXOff = 0; 
					 iXOff < iWidth - iDrawHeight; 
					 iXOff += iDrawHeight)
				{
					for (iYOff = 0; 
						 iYOff < iHeight - iDrawWidth; 
						 iYOff += iDrawWidth)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX, iCellY,
							 iCellWidth, iCellHeight);
					}

					if (iHeight - iYOff > 0)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX + (DrawFlags & Image2D::MirrorH ? 
									  iCellWidth - iHeight + iYOff : 0),
							 iCellY,
							 iHeight - iYOff,
							 iDrawHeight);
					}
				}

				if (iWidth - iXOff > 0)
				{
					for (iYOff = 0; 
						 iYOff < iHeight - iDrawWidth; 
						 iYOff += iDrawWidth)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX,
							 iCellY + (DrawFlags & Image2D::MirrorV ? 0 :
									  iCellHeight - iWidth + iXOff),
							 iDrawWidth,
							 iWidth - iXOff);
					}

					if (iHeight - iYOff > 0)
					{
						Draw(iX + iXOff, iY + iYOff,
							 0, 0,
							 (Image2D::Flags)(DrawFlags & ~Image2D::TileExtra),
							 modColor,
							 iCellX + (DrawFlags & Image2D::MirrorH ? 
									  iCellWidth - iHeight + iYOff : 0),
							 iCellY + (DrawFlags & Image2D::MirrorV ? 0 :
									  iCellHeight - iWidth + iXOff),
							 iHeight - iYOff,
							 iWidth - iXOff);
					}
				}
			}

			return;
		}
	}

	if (DrawFlags & Image2D::Rot90CW)
	{
		Draw(iX,iY,iHeight,iWidth,
			(Image2D::Flags)((DrawFlags & ~Image2D::Rot90CW)
			  | Image2D::RotTex),modColor,
			iCellX,iCellY,iCellWidth,iCellHeight);

		return;
	}

	int idX, idY;

	if ((iWidth != iDrawWidth || 
		iHeight != iDrawHeight) &&
		!(DrawFlags & Image2D::TileExtra))
	{
		idX = iX + iCellWidth*iWidth/iDrawWidth;
		idY = iY + iCellHeight*iHeight/iDrawHeight;
	}
	else
	{
		idX = iX + iCellWidth;
		idY = iY + iCellHeight;
	}

	float fWidthScale = (float)iWidth/(float)iDrawWidth;
	float fHeightScale = (float)iHeight/(float)iDrawHeight;

	float fU = (float)iCellX / (float)iTexWidth;
	float fV = (float)iCellY / (float)iTexHeight;
	float fdU = ((float)iCellX + (float)iCellWidth) / (float)iTexWidth;
	float fdV = ((float)iCellY + (float)iCellHeight) / (float)iTexHeight;

	if (DrawFlags & Image2D::MirrorH)
		std::swap(fU,fdU);

	if (DrawFlags & Image2D::MirrorV)
		std::swap(fV,fdV);

	if (iX < 0)
	{
		float f = -(float)iX / ((float)iTexWidth*fWidthScale);

		if (DrawFlags & Image2D::RotTex)
		{
			if (DrawFlags & Image2D::MirrorV)
				f *= -1;
			
			fdV -= f*iDrawHeight/iDrawWidth;
		}
		else
		{
			if (DrawFlags & Image2D::MirrorH)
				f *= -1;

			fU += f;
		}

		iX = 0;
	}

	if (idX >= iGEWidth)
	{
		float f = ((float)idX - (float)iGEWidth) / 
				((float)iTexWidth*fWidthScale);

		if (DrawFlags & Image2D::RotTex)
		{
			if (DrawFlags & Image2D::MirrorV)
				f *= -1;
			
			fV += f*(float)iDrawHeight/(float)iDrawWidth;
		}
		else
		{
			if (DrawFlags & Image2D::MirrorH)
				f *= -1;

			fdU -= f;
		}

		idX = iGEWidth - 1;
	}

	if (iY < 0)
	{
		float f = -(float)iY / ((float)iTexHeight*fHeightScale);

		if (DrawFlags & Image2D::RotTex)
		{
			if (DrawFlags & Image2D::MirrorH)
				f *= -1;
			
			fU += f*(float)iDrawWidth/(float)iDrawHeight;
		}
		else
		{
			if (DrawFlags & Image2D::MirrorV)
				f *= -1;

			fV += f;
		}

		iY = 0;
	}

	if (idY >= iGEHeight)
	{
		float f = ((float)idY - (float)iGEHeight) / 
				  ((float)iTexHeight*fHeightScale);

		if (DrawFlags & Image2D::RotTex)
		{
			if (DrawFlags & Image2D::MirrorH)
				f *= -1;
			
			fdU -= f*(float)iDrawWidth/(float)iDrawHeight;
		}
		else
		{
			if (DrawFlags & Image2D::MirrorV)
				f *= -1;

			fdV -= f;
		}

		idY = iGEHeight - 1;
	}

    glBegin(GL_TRIANGLE_STRIP);

	m_pGraphicsEngine->SetColor(modColor);

	if (DrawFlags & Image2D::RotTex)
	{
		glTexCoord2f(fU,fdV);
		glVertex2i(iX,iY);

		glTexCoord2f(fdU,fdV);
		glVertex2i(iX,idY);

		glTexCoord2f(fU,fV);
		glVertex2i(idX,iY);

		glTexCoord2f(fdU,fV);
		glVertex2i(idX,idY);
	}
	else
	{
		glTexCoord2f(fU,fV);
		glVertex2i(iX,iY);

		glTexCoord2f(fU,fdV);
		glVertex2i(iX,idY);

		glTexCoord2f(fdU,fV);
		glVertex2i(idX,iY);

		glTexCoord2f(fdU,fdV);
		glVertex2i(idX,idY);
	}

    glEnd();
}

unsigned int Texture::GetData(int x, int y)
{
	if (NULL == m_pData)
		throw IsisException("Uninitialized in GetData()");

    int offset = 4*(x + m_iHeight*y);

	return m_pData[offset+0] << 0 | 
           m_pData[offset+1] << 8 | 
           m_pData[offset+2] << 16 | 
           m_pData[offset+3] << 24;
}

} // namespace Isis
