
#include <isis/util.h>
#include <isis/ivm.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/Texture.h>

namespace Isis {

void Texture::FillExportTable()
{
    m_ExportTable["Init"]            = FUNC_INIT;
    m_ExportTable["LoadVRAM"]        = FUNC_LOADVRAM;
    m_ExportTable["UnloadVRAM"]      = FUNC_UNLOADVRAM;
    m_ExportTable["GetID"]           = FUNC_GETID;
    m_ExportTable["Draw"]            = FUNC_DRAW;
    m_ExportTable["Draw3D"]          = FUNC_DRAW3D;
    m_ExportTable["GetWidth"]        = FUNC_GETWIDTH;
    m_ExportTable["GetHeight"]       = FUNC_GETHEIGHT;
}

void Texture::IVMFillExportTable(std::vector<std::string> & table)
{
    std::map<std::string,FuncType>::iterator iter;

    for (iter = m_ExportTable.begin(); iter != m_ExportTable.end(); iter++)
    {
        table.push_back(iter->first);
    }
}

bool Texture::IVMContainsMethod(const std::string & funcname)
{
    return m_ExportTable.end() != m_ExportTable.find(funcname);
}

IVM::Value Texture::IVMCallMethod(const std::string & funcname,
                           const std::vector<IVM::Value> & parameters)
{   
    if (m_ExportTable.end() == m_ExportTable.find(funcname))
        throw IVM::IVMException("function not found in Texture object");
    
    switch (m_ExportTable[funcname])
    {
    case FUNC_INIT:
        {
            if (parameters.size() != 2)
                throw IVM::IVMException(
                       "wrong number of parameters for Load");

			IVM::IVMVector * pVector = dynamic_cast<IVM::IVMVector*>
									(parameters[1].GetObjPtr());

			if (NULL == pVector)
				throw IVM::IVMException("NULL vector for Load()");

			std::vector<Image2D*> stlVector;

            Init(parameters[0].GetString(),stlVector);

			for (int i = 0; i < (int)stlVector.size(); i++)
				pVector->Add(stlVector[i]);

        } break;
    case FUNC_LOADVRAM:
        {
            if (parameters.size() != 0)
                throw IVM::IVMException(
                         "wrong number of parameters for LoadVRAM");

            LoadVRAM();

            return true;

        } break;
    case FUNC_UNLOADVRAM:
        {
            if (parameters.size() != 0)
                throw IVM::IVMException(
                         "wrong number of parameters for UnloadVRAM");

            UnloadVRAM();

            return true;

        } break;
    case FUNC_GETID:
        {
            if (parameters.size() != 0)
                throw IVM::IVMException(
                         "wrong number of parameters for GetID");

            return (int)GetID();

        } break;
    case FUNC_GETWIDTH:
        {
            if (parameters.size() != 0)
                throw IVM::IVMException(
                         "wrong number of parameters for GetWidth");

            return GetWidth();

        } break;
    case FUNC_GETHEIGHT:
        {
            if (parameters.size() != 0)
                throw IVM::IVMException(
                         "wrong number of parameters for GetHeight");

            return GetHeight();

        } break;
    case FUNC_DRAW:
        {
			switch (parameters.size())
			{
			case 0:
				{
					Draw();
				} break;
			case 2:
				{
					Draw(parameters[0].GetInteger(),
						 parameters[1].GetInteger());
				} break;
			case 4:
				{
					Draw(parameters[0].GetInteger(),parameters[1].GetInteger(),
						parameters[2].GetInteger(),parameters[3].GetInteger());
				} break;
			case 5:
				{
					Draw(parameters[0].GetInteger(),parameters[1].GetInteger(),
						parameters[2].GetInteger(),parameters[3].GetInteger(),
						(Image2D::Flags)parameters[4].GetInteger());
				} break;
			case 6:
				{
					Draw(parameters[0].GetInteger(),parameters[1].GetInteger(),
						parameters[2].GetInteger(),parameters[3].GetInteger(),
						(Image2D::Flags)parameters[4].GetInteger(),
						parameters[5].GetInteger());
				} break;
			case 10:
				{
					Draw(parameters[0].GetInteger(),parameters[1].GetInteger(),
						parameters[2].GetInteger(),parameters[3].GetInteger(),
						(Image2D::Flags)parameters[4].GetInteger(),
						parameters[5].GetInteger(),
						parameters[6].GetInteger(),parameters[7].GetInteger(),
						parameters[8].GetInteger(),parameters[9].GetInteger());
				} break;
			default:
				{
					throw IVM::IVMException(
							"wrong number of parameters for Draw");
				} break;
			}

            return true;

        } break;
    case FUNC_DRAW3D:
        {
			switch (parameters.size())
			{
			case 0:
				{
					Draw3D();
				} break;
			case 2:
				{
					Draw3D(parameters[0].GetInteger(),
						   parameters[1].GetInteger());
				} break;
			case 4:
				{
					Draw3D(parameters[0].GetInteger(),parameters[1].GetInteger(),
						   parameters[2].GetInteger(),parameters[3].GetInteger());
				} break;
			case 5:
				{
					Draw3D(parameters[0].GetInteger(),parameters[1].GetInteger(),
						   parameters[2].GetInteger(),parameters[3].GetInteger(),
						   (Image2D::Flags)parameters[4].GetInteger());
				} break;
			case 6:
				{
					Draw3D(parameters[0].GetInteger(),parameters[1].GetInteger(),
						   parameters[2].GetInteger(),parameters[3].GetInteger(),
						   (Image2D::Flags)parameters[4].GetInteger(),
						   parameters[5].GetInteger());
				} break;
			case 10:
				{
					Draw3D(parameters[0].GetInteger(),parameters[1].GetInteger(),
						   parameters[2].GetInteger(),parameters[3].GetInteger(),
						   (Image2D::Flags)parameters[4].GetInteger(),
						   parameters[5].GetInteger(),
						   parameters[6].GetInteger(),parameters[7].GetInteger(),
						   parameters[8].GetInteger(),parameters[9].GetInteger());
				} break;
			default:
				{
					throw IVM::IVMException(
							"wrong number of parameters for Draw3D");
				} break;
			}

            return true;

        } break;
    }

	return true;
}

} // namespace Isis

