
#include <math.h>
#include <sys/stat.h>
#include <zlib.h>
#include <stdio.h>
#include <stdlib.h>

#include <algorithm>

#include <isis/util.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/VectorAnimation.h>
#include <isis/game/VectorImage.h>

namespace Isis {

#ifdef _DEBUG
int VectorAnimation::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,VectorAnimation::FuncType> VectorAnimation::m_ExportTable;

VectorAnimation::VectorAnimation()
{
    if (m_ExportTable.empty())
        FillExportTable();

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

VectorAnimation::~VectorAnimation()
{
	for (int i = 0; i < (int)m_Frames.size(); i++)
		m_Frames[i]->Release();

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int VectorAnimation::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void VectorAnimation::LoadVGZ(GraphicsEngine * pGraphicsEngine,
							  const std::string & sImageFilename,
							  std::vector<VectorAnimation*> & animations)
{
    struct stat statbuf;
	std::string prefix;
	bool bFound = false;
	std::vector<std::string> Prefixes = pGraphicsEngine->GetImagePrefixes();

	for (unsigned int i = 0; i < Prefixes.size(); i++)
	{
		if (-1 != stat((prefix + sImageFilename).c_str(),&statbuf))
		{
			bFound = true;
			break;
		}
	}

	if (!bFound)
		throw IsisException("Could not stat image file:",
							sImageFilename.c_str());

	gzFile file = gzopen((prefix + sImageFilename).c_str(),"rb");
	
	if (NULL == file)
		throw IsisException("Could not open image file");

	int count = ReadInt(file);

	for (int i = 0; i < count; i++)
	{
		VectorAnimation * pAnim = new VectorAnimation();
		pAnim->Init(pGraphicsEngine,file);
		animations.push_back(pAnim);
	}

	gzclose(file);
}

void VectorAnimation::Init(GraphicsEngine * pGraphicsEngine, gzFile file)
{
	m_sName = ReadString(file);
	m_fFramerate = ReadFloat(file);
	
	// syncframes
	ReadInt(file);

	int count = ReadInt(file);

	for (int i = 0; i < count; i++)
	{
		VectorImage * pImage = new VectorImage(pGraphicsEngine);
		pImage->Init(file);
		pImage->AddRef();
		m_Frames.push_back(pImage);
	}
}


int VectorAnimation::ReadInt(gzFile file)
{
	unsigned char bytes[4];

	if (4 != gzread(file,bytes,4))
	{
		gzclose(file);
		throw IsisException("Error reading image file");
	}

	return bytes[0] | (bytes[1] << 8) | 
		   (bytes[2] << 16) | (bytes[3] << 24);
}

float VectorAnimation::ReadFloat(gzFile file)
{
	int i = ReadInt(file);
	return *((float*)(&i));
}

std::string VectorAnimation::ReadString(gzFile file)
{
	int size = ReadInt(file);

	char * bytes = new char [size+1];
	bytes[size] = 0;

	if (size != gzread(file,bytes,size))
	{
		gzclose(file);
		delete [] bytes;
		throw IsisException("Error reading image file");
	}

	std::string string(bytes);

	delete [] bytes;

	return string;
}

} // namespace Isis
