
#include <math.h>
#include <sys/stat.h>
#include <zlib.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include <algorithm>

#include <isis/util.h>

#include <isis/game/GraphicsEngine.h>
#include <isis/game/VectorImage.h>
#include <isis/game/CellImage.h>

// dumb windows stuff
#ifndef _WIN32
#define CALLBACK
#define __stdcall
#endif

namespace Isis {

#ifdef _DEBUG
int VectorImage::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,VectorImage::FuncType> VectorImage::m_ExportTable;

VectorImage::VectorImage(GraphicsEngine * pGraphicsEngine)
{
    if (m_ExportTable.empty())
        FillExportTable();

    m_uiID = 0;

	m_pGraphicsEngine = pGraphicsEngine;
	m_pGraphicsEngine->AddRef();
    m_pGraphicsEngine->VRAMObjectCreated(this);

    m_fLeft = 0;
	m_fRight = 0;
	m_fTop = 0;
	m_fBottom = 0;

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

VectorImage::~VectorImage()
{
	UnloadVRAM();

	if (m_pGraphicsEngine)
    {
        m_pGraphicsEngine->VRAMObjectDestroyed(this);
		m_pGraphicsEngine->Release();
    }

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int VectorImage::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void VectorImage::Init(gzFile file)
{
    if (m_Records.size() > 0)
        throw IsisException("Already initialized in Init");

	// center x,y
	ReadInt(file);
	ReadInt(file);

	// bounds
	ReadInt(file);
	ReadInt(file);
	ReadInt(file);
	ReadInt(file);

	// zoom
	ReadInt(file);

	int length = ReadInt(file);
	gzseek(file,length,SEEK_CUR);

	m_fLeft = ReadFloat(file);
	m_fRight = ReadFloat(file);
	m_fTop = ReadFloat(file);
	m_fBottom = ReadFloat(file);

	RecordType type = (RecordType)ReadInt(file);

	int i = 0;
	while (type != EndRecord)
	{
		Record r;
		r.type = type;

		switch (type)
		{
		case ColorRecord:
			{
				Color c;
				c.red = ReadByte(file);
				c.green = ReadByte(file);
				c.blue = ReadByte(file);

				r.index = (int)m_Colors.size();
				m_Colors.push_back(c);

			} break;
		case PolyRecord:
			{
				Polygon poly;

				int num = ReadInt(file);

				if (num > 1000000 || num < 0)
					throw IsisException("Invalid number of points in polygon");

				for (int j = 0; j < num; j++)
				{
					Point point;
					point.x = ReadFloat(file);
					point.y = ReadFloat(file);

					poly.points.push_back(point);
				}

				r.index = (int)m_Polygons.size();
				m_Polygons.push_back(poly);

			} break;
		case PolyPolyRecord:
			{
				PolyPolygon polypoly;

				int num = ReadInt(file);

				if (num > 1000000 || num < 0)
					throw IsisException("Invalid number of polygons");

				for (int j = 0; j < num; j++)
				{
					Polygon poly;

					int points = ReadInt(file);

					if (points > 1000000 || points < 0)
						throw IsisException("Invalid number of points in polygon");

					for (int k = 0; k < points; k++)
					{
						Point point;
						point.x = ReadFloat(file);
						point.y = ReadFloat(file);

						poly.points.push_back(point);
					}

					polypoly.polygons.push_back(poly);
				}

				r.index = (int)m_PolyPolygons.size();
				m_PolyPolygons.push_back(polypoly);
			} break;
		default:
			{
		        throw IsisException("Unknown RecordType");
			}
		}

		m_Records.push_back(r);

		i++;
		type = (RecordType)ReadInt(file);
	}

	LoadVRAM();
}

static void __stdcall errorCallback(GLenum errorCode)
{
	const GLubyte * estring;
	estring = gluErrorString(errorCode);
	fprintf(stderr,"Tesselation error: %s\n",estring);
}

static std::vector<double*> mems;

static void CALLBACK combineCallback(GLdouble coords[3], GLdouble *vertex_data[4],
									GLfloat weight[4], GLdouble **dataOut)
{
	GLdouble *vertex;

	vertex = (GLdouble *) malloc(3*sizeof(GLdouble));
	vertex[0] = coords[0];
	vertex[1] = coords[1];
	vertex[2] = coords[2];

	*dataOut = vertex;

	mems.push_back(vertex);
}

void VectorImage::LoadVRAM()
{
    if (m_uiID != 0)
        throw IsisException("Already loaded in LoadVRAM()");

	m_uiID = glGenLists(1);
	glNewList(m_uiID,GL_COMPILE);

	GLUtesselator * tess = gluNewTess();
#ifdef _WIN32
	gluTessCallback(tess,GLU_TESS_ERROR,(GLvoid (__stdcall*) ())errorCallback);
	gluTessCallback(tess,GLU_TESS_BEGIN,(GLvoid (__stdcall*) ())glBegin);
	gluTessCallback(tess,GLU_TESS_END,(GLvoid (__stdcall*) ())glEnd);
	gluTessCallback(tess,GLU_TESS_VERTEX,(GLvoid (__stdcall*) ())glVertex3dv);
	gluTessCallback(tess,GLU_TESS_COMBINE,(GLvoid (__stdcall*) ())combineCallback);
#elif MACOSX
	gluTessCallback(tess,GLU_TESS_ERROR,(GLvoid (*) (...))errorCallback);
	gluTessCallback(tess,GLU_TESS_BEGIN,(GLvoid (*) (...))glBegin);
	gluTessCallback(tess,GLU_TESS_END,(GLvoid (*) (...))glEnd);
	gluTessCallback(tess,GLU_TESS_VERTEX,(GLvoid (*) (...))glVertex3dv);
	gluTessCallback(tess,GLU_TESS_COMBINE,(GLvoid (*) (...))combineCallback);
#else
	gluTessCallback(tess,GLU_TESS_ERROR,(GLvoid (*) ())errorCallback);
	gluTessCallback(tess,GLU_TESS_BEGIN,(GLvoid (*) ())glBegin);
	gluTessCallback(tess,GLU_TESS_END,(GLvoid (*) ())glEnd);
	gluTessCallback(tess,GLU_TESS_VERTEX,(GLvoid (*) ())glVertex3dv);
	gluTessCallback(tess,GLU_TESS_COMBINE,(GLvoid (*) ())combineCallback);
#endif
	gluTessProperty(tess, GLU_TESS_WINDING_RULE, GLU_TESS_WINDING_ODD); 
//	gluTessProperty(tess, GLU_TESS_BOUNDARY_ONLY, 1); 

	for (int i = 0; i < (int)m_Records.size(); i++)
	{
		switch (m_Records[i].type)
		{
		case ColorRecord:
			{
				Color c = m_Colors[m_Records[i].index];
				glColor3ub(c.red,c.green,c.blue);
			} break;
		case PolyRecord:
			{
				gluTessBeginPolygon(tess,NULL);
				gluTessBeginContour(tess);

				Polygon * p = &m_Polygons[m_Records[i].index];
				std::vector<double> coords;
				coords.reserve(3*p->points.size());

				for (int j = 0; j < (int)p->points.size()-1; j++)
//				for (int j = (int)p->points.size()-2; j >= 0; j--)
				{
					coords.push_back(p->points[j].x);
					coords.push_back(p->points[j].y);
					coords.push_back(0);

					gluTessVertex(tess,&coords[j*3],&coords[j*3]);
				}

				gluTessEndContour(tess);
				gluTessEndPolygon(tess);
			} break;
		case PolyPolyRecord:
			{
				gluTessBeginPolygon(tess,NULL);

				PolyPolygon * pp = &m_PolyPolygons[m_Records[i].index];

				int size = 0;
				for (int j = 0; j < (int)pp->polygons.size(); j++)
					size += (int)(3*(pp->polygons[j].points.size()-1));

				std::vector<double> coords;
				coords.reserve(size);

				int offset = 0;
				for (int j = 0; j < (int)pp->polygons.size(); j++)
				{
					Polygon * p = &pp->polygons[j];

					gluTessBeginContour(tess);

					for (int k = 0; k < (int)p->points.size()-1; k++)
//					for (int k = (int)p->points.size()-2; k >= 0; k--)
					{
						Point * point = &p->points[k];
						coords.push_back(point->x);
						coords.push_back(point->y);
						coords.push_back(0);

						gluTessVertex(tess,&coords[offset],&coords[offset]);
						offset += 3;
					}

					gluTessEndContour(tess);
				}

				gluTessEndPolygon(tess);
			} break;
		}
	}

	gluDeleteTess(tess);

	for (int i = 0; i < (int)mems.size(); i++)
		free(mems[i]);
	mems.clear();

	glEndList();
}

void VectorImage::UnloadVRAM()
{
	if (m_uiID != 0)
    {
        glDeleteLists(m_uiID,1);
		m_uiID = 0;
    }
}

void VectorImage::Draw(float x, float y, float z, float angle, 
					   float scale_x, float scale_y)
{
	if (m_uiID == 0)
		throw IsisException("VectorImage not initialized in Draw()");

	m_pGraphicsEngine->EnableTexMap(false);
	m_pGraphicsEngine->SetBlend(false);

	glPushMatrix();

	glTranslatef(x,y,z);

	if (angle != 0)
		glRotatef(angle,0,0,1);

	if (scale_y == 0)
		scale_y = scale_x;

	if (scale_x != 1 || scale_y != 1)
		glScalef(scale_x,scale_y,1);

	glCallList(m_uiID);

	glPopMatrix();
}

} // namespace Isis
