
#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#include <windows.h>
#endif

#include <isis/util.h>

#include "ISEReader.h"

#include <isis/ivm/IVMException.h>
#include <isis/ivm/isc.h>
#include <isis/ivm/isasm.h>

namespace IVM {

ISEReader::ISEReader()
{
    m_bFileOpen = false;
    m_pFile = NULL;
}

ISEReader::~ISEReader()
{
    if (m_bFileOpen)
    {
        CloseFile();
    }
}


void ISEReader::OpenFile(const std::string & filename)
{
    std::string newfilename = filename;

    if (m_bFileOpen)
    {
        CloseFile();
    }

    if (std::string::npos == filename.rfind(".isz"))
    {
        if (std::string::npos != filename.rfind(".ism"))
        {
            ConvertISM(filename,newfilename);

            tmpfile = newfilename;
        }
        else if (std::string::npos != filename.rfind(".iss"))
        {
            ConvertISS(filename,newfilename);

            tmpfile = newfilename;
        }
        else
            throw IVMException("file is in unknown format");
    }

    m_pFile = gzopen(newfilename.c_str(),"rb");

    if (NULL == m_pFile)
		throw IVMException("fopen() failed for file: ",newfilename);
    
    m_bFileOpen = true;
}

void ISEReader::CloseFile()
{
    if (NULL != m_pFile)
    {
        gzclose(m_pFile);
        m_pFile = NULL;
    }

    m_bFileOpen = false;

    if (tmpfile.length() > 0)
    {
        if (-1 == remove(tmpfile.c_str()))
            throw IVMException("error removing temporary file");
    }
}

void ISEReader::ConvertISM(const std::string & oldfile, std::string & newfile)
{
#ifdef _DEBUG
	printf("assembling file: %s ... ",oldfile.c_str());

# ifdef WIN32
	SetEnvironmentVariable("ISISSCRIPT_DEBUG","1");
# else
	setenv("ISISSCRIPT_DEBUG","1",1);
# endif // WIN32

#endif // _DEBUG

//#ifdef MACOSX
//    if (0 != system(std::string("bin/isasm \"" + oldfile + "\"").c_str()))
//        throw IVMException("error assembling file");
//#else
//    if (0 != system(std::string("isasm \"" + oldfile + "\"").c_str()))
//        throw IVMException("error assembling file");
//#endif

    isasm_reset_global_state();

    ISASM_in = fopen(oldfile.c_str(),"rt");

    if (NULL == ISASM_in)
        throw IVMException("Error opening temporary file: ",oldfile);

    int result = ISASM_parse();
    fclose(ISASM_in);

    if (0 != result)
        throw IVMException("Error parsing file: ",oldfile);

    newfile = oldfile.substr(0,oldfile.length()-4) + ".isz";

    gzFile outfile = gzopen(newfile.c_str(),"wb");

    result = save_isz_file(outfile);
    gzclose(outfile);

    if (0 == result)
        throw IVMException("Error saving output file: ",newfile);
}

void ISEReader::ConvertISS(const std::string & oldfile, std::string & newfile)
{
#ifdef _DEBUG
	printf("compiling file: %s ... ",oldfile.c_str());

# ifdef WIN32
	SetEnvironmentVariable("ISISSCRIPT_DEBUG","1");
# else
	setenv("ISISSCRIPT_DEBUG","1",1);
# endif // WIN32

#endif // _DEBUG

	std::string tmpname;

//#ifdef MACOSX
//    if (0 != system(std::string("bin/isc \"" + oldfile + "\"").c_str()))
//        throw IVMException("error compiling file");
//#else
//    if (0 != system(std::string("isc \"" + oldfile + "\"").c_str()))
//        throw IVMException("error compiling file");
//#endif

    isc_reset_global_state();

    ISC_in = fopen(oldfile.c_str(),"rt");
    g_sCurFile = oldfile;

    int result = ISC_parse();
    fclose(ISC_in);

    if (0 != result)
		throw IVMException("Error compiling file: ",oldfile);

    tmpname = oldfile.substr(0,oldfile.length()-4) + ".ism";

    FILE * outfile = fopen(tmpname.c_str(),"wt");

    if (NULL == outfile)
        throw IVMException("Error opening output file: ",tmpname);

    result = save_ism_file(oldfile.c_str(),tmpname.c_str(),outfile);
    fclose(outfile);

    if (0 == result)
        throw IVMException("Error saving output file: ",tmpname);

//#ifdef MACOSX
//    if (0 != system(std::string("bin/isasm \"" + tmpname + "\"").c_str()))
//        throw IVMException("error assembling file");
//#else
//    if (0 != system(std::string("isasm \"" + tmpname + "\"").c_str()))
//        throw IVMException("error assembling file");
//#endif

    isasm_reset_global_state();

    ISASM_in = fopen(tmpname.c_str(),"rt");

    if (NULL == ISASM_in)
        throw IVMException("Error opening temporary file: ",tmpname);

    result = ISASM_parse();
    fclose(ISASM_in);

    if (0 != result)
        throw IVMException("Error parsing file: ",tmpname);

    newfile = tmpname.substr(0,tmpname.length()-4) + ".isz";

    gzFile outfile2 = gzopen(newfile.c_str(),"wb");

    result = save_isz_file(outfile2);
    gzclose(outfile2);

    if (0 == result)
        throw IVMException("Error saving output file: ",newfile);

#ifdef _DEBUG
	printf("done.\n");
#endif // _DEBUG

    if (-1 == remove(tmpname.c_str()))
        throw IVMException("error removing temporary file");
}

void ISEReader::CheckID()
{
    size_t br;
    unsigned char id[4];

    if (!m_bFileOpen)
        throw IVMException("file is not open in CheckID");

    br = gzread(m_pFile,id,sizeof(unsigned char)*4);
    
    if ((4*sizeof(unsigned char)) != br)
        throw IVMException("fread() error 1");

    if (!((id[0] == 'I') && 
          (id[1] == 'S') && 
          (id[2] == 'E') && 
          (id[3] == 'F')))
    {
        throw IVMException("file ID does not match ISEF");
    }
}

void ISEReader::ReadMajorMinor(int & major, int & minor)
{
    if (!m_bFileOpen)
        throw IVMException("file is not open in ReadMajorMinor");

    major = ReadByte();
    minor = ReadByte();
}

int ISEReader::ReadInt()
{
    size_t br;
    unsigned char bytes[4];

    if (!m_bFileOpen)
        throw IVMException("file is not open in ReadInt");

    br = gzread(m_pFile,bytes,sizeof(unsigned char)*4);
   
    if ((4*sizeof(unsigned char)) != br)
        throw IVMException("fread() error 2");

    return ((bytes[0] << 0) |
           (bytes[1] << 8) | 
           (bytes[2] << 16) | 
           (bytes[3] << 24));

}

float ISEReader::ReadFloat()
{
    int temp = ReadInt();
    return *((float*)(&temp));
}

std::string ISEReader::ReadString()
{
    size_t br;
    int length;
    char * string;
    std::string value;

    if (!m_bFileOpen)
        throw IVMException("file is not open");

	length = ReadInt();

    if (length < 0 || length >= (1024*1024)) // 1 megabyte limit on strings
        throw IVMException("invalid string size");

    string = (char *)malloc(length+1);

    if (NULL == string)
        throw IVMException("memory allocation failure");

    br = gzread(m_pFile,string,sizeof(char)*length);

    if ((length*sizeof(char)) != br)
    {
        free(string);
        throw IVMException("fread() failure");
    }

    string[length] = '\0';

	value = string;

    free(string);

    return value;
}

unsigned char ISEReader::ReadByte()
{
    size_t br;
    unsigned char value;

    if (!m_bFileOpen)
        throw IVMException("file is not open");

    br = gzread(m_pFile,&value,sizeof(unsigned char)*1);

    if (sizeof(unsigned char) != br)
        throw IVMException("fread() failure");

    return value;
}

} // namespace ivm

