
#include <math.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>

#include <isis/ivm/IVMException.h>
#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Script.h>
#include <isis/ivm/IVMVector.h>
#include <isis/ivm/DataSet.h>

#include "IVMStdLibProvider.h"
#include "IVMConsole.h"

namespace IVM {

std::map<std::string,IVMStdLibProvider::ObjectType> 
                                         IVMStdLibProvider::m_ObjectTable;

std::map<std::string,IVMStdLibProvider::FuncType> 
                                         IVMStdLibProvider::m_ExportTable;

IVMStdLibProvider::IVMStdLibProvider()
{
    if (m_ObjectTable.empty())
        FillObjectTable();

	if (m_ExportTable.empty())
		FillExportTable();
}

void IVMStdLibProvider::FillObjectTable()
{
    m_ObjectTable["Console"] = OBJECT_CONSOLE;
    m_ObjectTable["IsisVM"]  = OBJECT_ISISVM;
    m_ObjectTable["Vector"]  = OBJECT_VECTOR;
    m_ObjectTable["DataSet"] = OBJECT_DATASET;
}

void IVMStdLibProvider::FillObjectTable(std::vector<std::string> & table)
{
    std::map<std::string,ObjectType>::iterator iter;

    for (iter = m_ObjectTable.begin(); iter != m_ObjectTable.end(); iter++)
    {
        table.push_back(iter->first);
    }
}

IVMObject * IVMStdLibProvider::CreateObject(const std::string & objname,
                                         const std::vector<Value> & InitParams)
{
    if (m_ObjectTable.end() == m_ObjectTable.find(objname))
        throw IVMException("object type not found");

    IVMObject * pObject;

    switch (m_ObjectTable[objname])
    {
    case OBJECT_CONSOLE:
        {
            pObject = new IVMConsole();
        } break;
    case OBJECT_ISISVM:
        {
            pObject = new IsisVM();
        } break;
    case OBJECT_VECTOR:
        {
            pObject = new IVMVector();
        } break;
    case OBJECT_DATASET:
        {
            pObject = new DataSet();
        } break;
    default:
        {
			throw IVMException("object type not implemented: ",objname);
        }
    }

	pObject->AddRef();

    if (pObject->IVMContainsMethod("Init"))
    {
        pObject->IVMCallMethod("Init",InitParams);
    }
    else if (!InitParams.empty())
    {
        throw IVMException("Init parameters passed but no Init() was found");
    }

	pObject->SafeRelease();
    return pObject;
}

void IVMStdLibProvider::FillExportTable()
{
    m_ExportTable["ChopCharacter"] = FUNC_CHOPCHARACTER;
    m_ExportTable["ToUpper"]       = FUNC_TOUPPER;
    m_ExportTable["ToLower"]       = FUNC_TOLOWER;
	m_ExportTable["StrLen"]        = FUNC_STRLEN;
    m_ExportTable["IsNull"]        = FUNC_ISNULL;
    m_ExportTable["Int"]           = FUNC_INT;
    m_ExportTable["Float"]         = FUNC_FLOAT;
    m_ExportTable["String"]        = FUNC_STRING;
    m_ExportTable["ListFiles"]     = FUNC_LISTFILES;
    m_ExportTable["Rand"]          = FUNC_RAND;
    m_ExportTable["sin"]           = FUNC_SIN;
    m_ExportTable["cos"]           = FUNC_COS;
    m_ExportTable["sqrt"]          = FUNC_SQRT;
    m_ExportTable["pow"]           = FUNC_POW;
    m_ExportTable["abs"]           = FUNC_ABS;
}

void IVMStdLibProvider::IVMFillExportTable(std::vector<std::string> & table)
{
    std::map<std::string,FuncType>::iterator iter;

    for (iter = m_ExportTable.begin(); iter != m_ExportTable.end(); iter++)
    {
        table.push_back(iter->first);
    }
}

bool IVMStdLibProvider::IVMContainsMethod(const std::string & funcname)
{
    return m_ExportTable.end() != m_ExportTable.find(funcname);
}

Value IVMStdLibProvider::IVMCallMethod(const std::string & funcname,
                           const std::vector<IVM::Value> & parameters)
{   
    if (m_ExportTable.end() == m_ExportTable.find(funcname))
        throw IVM::IVMException(
                "function not found in IVMStdLibProvider object");
    
    switch (m_ExportTable[funcname])
    {
    case FUNC_CHOPCHARACTER:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for ChopCharacter");

			std::string oldstring = parameters[0].GetString();
			std::string newstring;
			newstring.assign(oldstring.begin(),oldstring.end()-1);

            return newstring;

        } break;
    case FUNC_TOUPPER:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for ToUpper");

			std::string oldstring = parameters[0].GetString();

			for (unsigned int i = 0; i < oldstring.length(); i++)
			{
				char c = oldstring[i];

				if (isalpha(c))
					c = (char)toupper(c);
				else
				{
					switch (c)
					{
					case '`': c = '~'; break;
					case '1': c = '!'; break;
					case '2': c = '@'; break;
					case '3': c = '#'; break;
					case '4': c = '$'; break;
					case '5': c = '%'; break;
					case '6': c = '^'; break;
					case '7': c = '&'; break;
					case '8': c = '*'; break;
					case '9': c = '('; break;
					case '0': c = ')'; break;
					case '-': c = '_'; break;
					case '=': c = '+'; break;
					case '[': c = '{'; break;
					case ']': c = '}'; break;
					case '\\': c = '|'; break;
					case ';': c = ':'; break;
					case '\'': c = '\"'; break;
					case ',': c = '<'; break;
					case '.': c = '>'; break;
					case '/': c = '?'; break;
					}
				}

				oldstring[i] = c;
			}

            return oldstring;

        } break;
    case FUNC_TOLOWER:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for ToLower");

			std::string oldstring = parameters[0].GetString();

			for (unsigned int i = 0; i < oldstring.length(); i++)
			{
				oldstring[i] = (char)tolower(oldstring[i]);
			}

            return oldstring;

        } break;
    case FUNC_STRLEN:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for StrLen");

            return (int)(parameters[0].GetString().length());

        } break;
    case FUNC_ISNULL:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for IsNull");

			switch (parameters[0].GetType())
			{
			case Value::None:
				{
					return true;
				} break;
			case Value::Integer:
			case Value::Float:
			case Value::String:
				{
					return parameters[0].GetInteger() == 0;
				} break;
			case Value::ObjPtr:
				{
					return parameters[0].GetObjPtr() == NULL;
				} break;
			}

        } break;
    case FUNC_INT:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for Int");

			return parameters[0].GetInteger();

        } break;
    case FUNC_FLOAT:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for Float");

			return parameters[0].GetFloat();

        } break;
    case FUNC_RAND:
        {
            switch (parameters.size())
            {
            case 0:
                {
                    return rand();
                } break;
            case 1:
                {
                    return rand() % parameters[0].GetInteger();
                } break;
            case 2:
                {
                    int min = parameters[0].GetInteger();
                    int max = parameters[1].GetInteger();

                    return min + rand() % (max-min);

                } break;
            default:
                {
                    throw IVM::IVMException(
                            "wrong number of parameters for Rand");
                }
            }
        } break;
    case FUNC_STRING:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for String");

			return parameters[0].GetString();

        } break;
	case FUNC_LISTFILES:
		{
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for ListFiles");

			DIR * dir = opendir(parameters[0].GetString().c_str());

			if (NULL == dir)
				throw IVM::IVMException("error opening directory:",
										parameters[0].GetString());

			IVM::IVMVector * pVector = new IVM::IVMVector();
			struct dirent * entry;

			while (NULL != (entry = readdir(dir)))
			{
				if (strcmp(".",entry->d_name) && strcmp("..",entry->d_name))
					pVector->Add(std::string(entry->d_name));
			}

			closedir(dir);

			return pVector;

		} break;
    case FUNC_SIN:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for sin");

			return (float)sin(parameters[0].GetFloat());

        } break;
    case FUNC_COS:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for cos");

			return (float)cos(parameters[0].GetFloat());

        } break;
    case FUNC_SQRT:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for sqrt");

			return (float)sqrt(parameters[0].GetFloat());

        } break;
    case FUNC_POW:
        {
            if (parameters.size() != 2)
                throw IVM::IVMException(
                            "wrong number of parameters for pow");

			return (float)pow(parameters[0].GetFloat(),parameters[1].GetFloat());

        } break;
    case FUNC_ABS:
        {
            if (parameters.size() != 1)
                throw IVM::IVMException(
                            "wrong number of parameters for abs");

			return (float)fabs(parameters[0].GetFloat());

        } break;
    }

	return true;
}

} // namespace IVM

