
#include <stdio.h>
#include <iostream>
#include <algorithm>

#include <isis/ivm/Script.h>
#include <isis/ivm/IVMException.h>
#include <isis/ivm/IVMVector.h>

namespace IVM {

#ifdef _DEBUG
int IVMVector::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,IVMVector::FuncType> IVMVector::m_ExportTable;

IVMVector::IVMVector()
{
    if (m_ExportTable.empty())
        FillExportTable();

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

IVMVector::~IVMVector()
{
#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int IVMVector::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void IVMVector::FillExportTable()
{
    m_ExportTable["Init"] =      FUNC_INIT;
    m_ExportTable["Add"] =       FUNC_ADD;
    m_ExportTable["Get"] =       FUNC_GET;
    m_ExportTable["Set"] =       FUNC_SET;
    m_ExportTable["DelValue"] =  FUNC_DELVALUE;
    m_ExportTable["DelIndex"] =  FUNC_DELINDEX;
    m_ExportTable["Clear"] =     FUNC_CLEAR;
    m_ExportTable["Size"] =      FUNC_SIZE;
    m_ExportTable["Save"] =      FUNC_SAVE;
    m_ExportTable["Load"] =      FUNC_LOAD;
}

bool IVMVector::IVMContainsMethod(const std::string & funcname)
{
    return m_ExportTable.end() != m_ExportTable.find(funcname);
}
 
void IVMVector::IVMFillExportTable(std::vector<std::string> & table)
{
    std::map<std::string,FuncType>::iterator iter;

    for (iter = m_ExportTable.begin(); iter != m_ExportTable.end(); iter++)
    {
        table.push_back(iter->first);
    }
}

void IVMVector::Add(Value value)
{
    m_Vector.push_back(value);
}

Value IVMVector::Get(int index)
{
    if (index < 0 || index >= (int)m_Vector.size())
        throw IVMException("IVMVector index out of bounds in Get");

    return m_Vector[index];
}

void IVMVector::Set(int index, Value val)
{
    if (index < 0 || index >= (int)m_Vector.size())
        throw IVMException("IVMVector index out of bounds in Get");

    m_Vector[index] = val;
}

void IVMVector::DelValue(Value value)
{
	m_Vector.erase(std::remove(m_Vector.begin(),m_Vector.end(),value),
                  m_Vector.end());
}

void IVMVector::DelIndex(int index)
{
	if (index < 0 || index >= (int)m_Vector.size())
		throw IVMException("index out of bounds in Vector::DelIndex");

	m_Vector.erase(m_Vector.begin() + index);
}

void IVMVector::Clear()
{
    m_Vector.clear();
}

int IVMVector::Size()
{
    return (int)m_Vector.size();
}

void IVMVector::Save(ISaveLoad * pSaveLoad)
{
	if (NULL == pSaveLoad)
		throw new IVMException("NULL ISaveLoad given in IVMVector::Save()");

    pSaveLoad->Save(Size());
    
    for (int i = 0; i < Size(); i++)
        pSaveLoad->Save(Get(i));
}

void IVMVector::Load(ISaveLoad * pSaveLoad)
{
    Clear();

    int size = pSaveLoad->Load().GetInteger();

    for (int i = 0; i < size; i++)
        Add(pSaveLoad->Load());
}

Value IVMVector::IVMCallMethod(const std::string & funcname, 
                               const std::vector<Value> & parameters)
{
    if (m_ExportTable.end() == m_ExportTable.find(funcname))
        throw IVMException("function not found in IVMVector",
							funcname);

    switch (m_ExportTable[funcname])
    {
    case FUNC_INIT:
        {
			if (!parameters.empty())
	            m_Vector.insert(m_Vector.end(),parameters.begin(),
                                parameters.end());

		} break;
    case FUNC_ADD:
        {
            if (parameters.size() == 0)
                throw IVMException("no parameters to Add");

            m_Vector.insert(m_Vector.end(),parameters.begin(),parameters.end());

            return true;
        } break;
    case FUNC_GET:
        {
            if (parameters.size() != 1)
                throw IVMException("wrong number of parameters for Get");

            return Get(parameters[0].GetInteger());
        } break;
    case FUNC_SET:
        {
            if (parameters.size() != 2)
                throw IVMException("wrong number of parameters for Set");

			Set(parameters[0].GetInteger(),parameters[1]);

            return true;
        } break;
    case FUNC_DELINDEX:
        {
            if (parameters.size() == 0)
                throw IVMException("no parameters to Del");

            for (unsigned int i = 0; i < parameters.size(); i++)
            {
				DelIndex(parameters[i].GetInteger());
            }

            return true;
        } break;
    case FUNC_DELVALUE:
        {
            if (parameters.size() == 0)
                throw IVMException("no parameters to Del");

            for (unsigned int i = 0; i < parameters.size(); i++)
            {
				DelValue(parameters[i]);
            }

            return true;
        } break;
    case FUNC_CLEAR:
        {
            if (parameters.size() != 0)
                throw IVMException("wrong number of parameters for Clear");

            m_Vector.clear();

            return true;
        } break;
    case FUNC_SIZE:
        {
            if (parameters.size() != 0)
                throw IVMException("wrong number of parameters for Size");

            return (int)m_Vector.size();
        } break;
    case FUNC_SAVE:
        {
            if (parameters.size() != 1)
                throw IVMException("wrong number of parameters for Save");

            Save(dynamic_cast<ISaveLoad*>(parameters[0].GetObjPtr()));

            return true;
        } break;
    case FUNC_LOAD:
        {
            if (parameters.size() != 1)
                throw IVMException("wrong number of parameters for Load");

            Load(dynamic_cast<ISaveLoad*>(parameters[0].GetObjPtr()));

            return true;
        } break;
    default:
        {
            throw IVMException("function not found in IVMVector");
        } break;
    }

	return true;
}

} // namespace IVM

