
#include <isis/util.h>

#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Script.h>
#include <isis/ivm/IVMException.h>
#include <isis/ivm/ObjectProvider.h>

#include "ISEStruct.h"
#include "IVMStdLibProvider.h"
#include "IVMConsole.h"

namespace IVM {

#ifdef _DEBUG
int IsisVM::m_iGlobalRefCount = 0;
#endif // _DEBUG

std::map<std::string,IsisVM::FuncType> IsisVM::m_ExportTable;

IsisVM::IsisVM()
{
	m_bDestroyed = false;

    if (m_ExportTable.empty())
        FillExportTable();

    RegisterHAPIProvider(this);
	RegisterHAPIProvider(new IVMConsole());
	IVMStdLibProvider * pStdLibProvider = new IVMStdLibProvider();
    RegisterObjectProvider(pStdLibProvider);
    RegisterHAPIProvider(pStdLibProvider);

	AddScriptPrefix("");

#ifdef _DEBUG
	m_iGlobalRefCount++;
#endif // _DEBUG
}

IsisVM::~IsisVM()
{
	if (!m_bDestroyed)
		throw IVMException("IsisVM not Destroyed() before deletion");

#ifdef _DEBUG
	m_iGlobalRefCount--;
#endif // _DEBUG
}

#ifdef _DEBUG
int IsisVM::GetGlobalRefCount()
{
	return m_iGlobalRefCount;
}
#endif // _DEBUG

void IsisVM::Destroy()
{
	if (m_bDestroyed)
		throw IVMException("Already destroyed in Destroy()");

	m_bDestroyed = true;

	RemoveAllHAPIProviders();
    RemoveAllObjectProviders();
}

void IsisVM::AddScriptPrefix(const std::string & sScriptPrefix)
{
	bool bFound = false;

	std::vector<std::string>::iterator iter;
	for (iter = m_ScriptPrefixes.begin();
		 iter != m_ScriptPrefixes.end();
		 iter++)
	{
		if (*iter == sScriptPrefix)
		{
			bFound = true;
			break;
		}
	}

	if (!bFound)
		m_ScriptPrefixes.push_back(sScriptPrefix);
}

void IsisVM::RemoveScriptPrefix(const std::string & sScriptPrefix)
{
	std::vector<std::string>::iterator iter;

	for (iter = m_ScriptPrefixes.begin();
		 iter != m_ScriptPrefixes.end();
		 iter++)
	{
		if (*iter == sScriptPrefix)
		{
			m_ScriptPrefixes.erase(iter);
			break;
		}
	}
}

Script * IsisVM::MakeScript(const std::string & sScriptFilename,
                            const std::vector<Value> & InitParams)
{
	Script * pScript = new Script();
	pScript->AddRef();

	pScript->Load(this,sScriptFilename);

    if (pScript->ContainsFunction("Init"))
    {
        pScript->ExecSync("Init",InitParams);
    }
    else if (!InitParams.empty())
    {
        throw IVMException(
                "Init parameters passed to MakeScript, but no Init() exists");
    }

	pScript->SafeRelease();
    return pScript;
}

void IsisVM::RegisterHAPIProvider(IVMObject * provider)
{
    unsigned int i;
    std::vector<std::string> funcs;

    if (NULL == provider)
        throw IVMException("NULL object passed to RegisterHAPIProvider");

    provider->AddRef();

    provider->IVMFillExportTable(funcs);

    for (i = 0; i < funcs.size(); i++)
    {
		if ("Init" == funcs[i])
			continue;

        if (m_HAPITable.end() != m_HAPITable.find(funcs[i]))
            throw IVMException("HAPI func is already registered",
							   funcs[i]);

        m_HAPITable[funcs[i]] = provider;
    }
}

void IsisVM::RemoveHAPIProvider(IVMObject * provider)
{
    HAPITable::iterator iter_loop;
    HAPITable::iterator iter_last;

    for (iter_loop = m_HAPITable.begin(); iter_loop != m_HAPITable.end(); )
    {
        iter_last = iter_loop++;

        if (iter_last->second == provider)
            m_HAPITable.erase(iter_last);
    }

    provider->Release();
}

void IsisVM::RemoveAllHAPIProviders()
{
    while (!m_HAPITable.empty())
        RemoveHAPIProvider(m_HAPITable.begin()->second);
}

bool IsisVM::HAPIProviderExists(const std::string & funcname)
{
    return (NULL != FindHAPIProvider(funcname));
}

IVMObject * IsisVM::FindHAPIProvider(const std::string & funcname)
{
    HAPITable::iterator iter = m_HAPITable.find(funcname);

    if (iter == m_HAPITable.end())
        return NULL;

    return iter->second;
}


void IsisVM::RegisterObjectProvider(ObjectProvider * provider)
{
    unsigned int i;
    std::vector<std::string> objs;

    if (NULL == provider)
        throw IVMException("NULL object passed to RegisterObjectProvider");

    provider->AddRef();

    provider->FillObjectTable(objs);

    for (i = 0; i < objs.size(); i++)
    {
        if (m_ObjectProviderTable.end() != m_ObjectProviderTable.find(objs[i]))
            throw IVMException("object is already registered");

        m_ObjectProviderTable[objs[i]] = provider;
    }
}

void IsisVM::RemoveObjectProvider(ObjectProvider * provider)
{
    ObjProviderTable::iterator iter_loop;
    ObjProviderTable::iterator iter_last;

    for (iter_loop = m_ObjectProviderTable.begin(); 
         iter_loop != m_ObjectProviderTable.end(); )
    {
        iter_last = iter_loop++;

        if (iter_last->second == provider)
            m_ObjectProviderTable.erase(iter_last);
    }

    provider->Release();
}

void IsisVM::RemoveAllObjectProviders()
{
    while (!m_ObjectProviderTable.empty())
        RemoveObjectProvider(m_ObjectProviderTable.begin()->second);
}

bool IsisVM::ObjectProviderExists(const std::string & funcname)
{
    return (NULL != FindObjectProvider(funcname));
}

ObjectProvider * IsisVM::FindObjectProvider(const std::string & funcname)
{
    ObjProviderTable::iterator iter = m_ObjectProviderTable.find(funcname);

    if (iter == m_ObjectProviderTable.end())
        return NULL;

    return iter->second;
}


Value IsisVM::CallAPI(const std::string & sFuncname, 
		              const std::vector<Value> & Params)
{
	IVMObject * pProvider = FindHAPIProvider(sFuncname);

	if (NULL == pProvider)
		throw IVMException("API function not found: ",sFuncname);

	Script * pScript = new IVM::Script();
	pScript->AddRef();
	std::vector<Value> parameters;
	pScript->SetRetVal(pProvider->IVMCallMethod(sFuncname,Params));

	Value retval = pScript->GetRetVal();

	pScript->Release();

	return retval;
}

} // namespace IVM

