
#include <algorithm>

#include <isis/util.h>

#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Script.h>
#include <isis/ivm/Value.h>
#include <isis/ivm/IVMException.h>
#include <isis/ivm/IVMVector.h>

namespace IVM {

void IsisVM::FillExportTable()
{
    m_ExportTable["AddScriptPrefix"]      = FUNC_ADDSCRIPTPREFIX;
    m_ExportTable["RemoveScriptPrefix"]   = FUNC_REMOVESCRIPTPREFIX;
    m_ExportTable["MakeScript"]           = FUNC_MAKESCRIPT;
    m_ExportTable["RegisterHAPIProvider"] = FUNC_REGISTERHAPIPROVIDER;
    m_ExportTable["RemoveHAPIProvider"]   = FUNC_REMOVEHAPIPROVIDER;
    m_ExportTable["HAPIProviderExists"]   = FUNC_HAPIPROVIDEREXISTS;
    m_ExportTable["FindHAPIProvider"]     = FUNC_FINDHAPIPROVIDER;
    m_ExportTable["GetVM"]                = FUNC_GETVM;
    m_ExportTable["New"]                  = FUNC_NEW;
}

bool IsisVM::IVMContainsMethod(const std::string & funcname)
{
    return m_ExportTable.end() != m_ExportTable.find(funcname);
}

void IsisVM::IVMFillExportTable(std::vector<std::string> & table)
{
    std::map<std::string,FuncType>::iterator iter;

    for (iter = m_ExportTable.begin(); iter != m_ExportTable.end(); iter++)
    {
        table.push_back(iter->first);
    }
}

Value IsisVM::IVMCallMethod(const std::string & funcname, 
                            const std::vector<Value> & parameters)
{
    if (m_ExportTable.end() == m_ExportTable.find(funcname))
        throw IVMException("function not found in IsisVM object");

    switch (m_ExportTable[funcname])
    {
    case FUNC_ADDSCRIPTPREFIX:
        {
            if (parameters.size() != 1)
                throw IVMException(
                     "incorrect number of parameters for AddScriptPrefix");

			AddScriptPrefix(parameters[0].GetString());

            return true;
        } break;
    case FUNC_REMOVESCRIPTPREFIX:
        {
            if (parameters.size() != 1)
                throw IVMException(
                     "incorrect number of parameters for RemoveScriptPrefix");

			RemoveScriptPrefix(parameters[0].GetString());

            return true;
        } break;
    case FUNC_MAKESCRIPT:
        {
            if (parameters.size() <= 1)
                throw IVMException(
                     "incorrect number of parameters for MakeScript");
			
            std::vector<Value> NewParams;
            NewParams.assign(parameters.begin()+1,parameters.end());

            return MakeScript(parameters[0].GetString(),
                                         NewParams);

        } break;
    case FUNC_REGISTERHAPIPROVIDER:
        {
            if (parameters.size() != 1)
                throw IVMException("incorrect number of parameters for "
                        "RegisterHAPIProvider");
            
            RegisterHAPIProvider(parameters[0].GetObjPtr());
            return true;

        } break;
    case FUNC_REMOVEHAPIPROVIDER:
        {
            if (parameters.size() != 1)
                throw IVMException("incorrect number of parameters for "
                        "RemoveHAPIProvider");

            RemoveHAPIProvider(parameters[0].GetObjPtr());
            return true;

        } break;
    case FUNC_HAPIPROVIDEREXISTS:
        {
            if (parameters.size() != 1)
                throw IVMException("incorrect number of parameters for "
                        "HAPIProviderExists");

            return HAPIProviderExists(parameters[0].GetString());

        } break;
    case FUNC_FINDHAPIPROVIDER:
        {
            if (parameters.size() != 1)
                throw IVMException("incorrect number of parameters for "
                        "FindHAPIProvider");

            IVMObject * provider = FindHAPIProvider(parameters[0].GetString());

            if (NULL == provider)
                throw IVMException("HAPI provider not found");

            return provider;

        } break;
    case FUNC_GETVM:
        {
            if (parameters.size() != 0)
                throw IVMException("incorrect number of parameters for GetVm");

            return this;

        } break;
    case FUNC_NEW:
        {
            if (parameters.size() == 0)
                throw IVMException("no parameters passed to new()!");

            std::vector<Value> NewParams;
            NewParams.assign(parameters.begin(),parameters.end()-1);

            IVMObject * object = NewObject(
                                   parameters[parameters.size()-1].GetString(),
                                   NewParams);

            return object;

        } break;
    default:
        {
            throw IVMException("function not found");
        }
    }
}

} // namespace IVM

