
#include <stdio.h>

#include <isis/util.h>

#include <isis/ivm/IVMException.h>
#include <isis/ivm/IsisVM.h>

#include "ISEStruct.h"
#include "ISEReader.h"

namespace IVM {

std::string IsisVM::NameOfISE(ISEStruct * pISE)
{
    std::map<std::string,ISEStruct *>::iterator iter;

	for (iter = m_ISEStructs.begin();
		 iter != m_ISEStructs.end();
		 iter++)
	{
		if (iter->second == pISE)
			return iter->first;
	}

	return "NULL";
}


ISEStruct * IsisVM::LoadISE(const std::string & filename)
{
    ISEReader reader;
    ISEStruct * out;
    int num;
    std::map<std::string,ISEStruct *>::iterator old;

    old = m_ISEStructs.find(filename);

    if (old != m_ISEStructs.end())
    {
        out = old->second;
        return out;
    }

    out = new ISEStruct();
	out->Filename = filename;

    reader.OpenFile(filename);
    //INFO_MSG("opened file: %s",filename.c_str());

    reader.CheckID();
    //INFO_MSG("ID check OK");

    reader.ReadMajorMinor(out->iMajorVersion,out->iMinorVersion);
    //INFO_MSG("versions  major: %d  minor: %d",
    //         out->iMajorVersion,out->iMinorVersion);

    out->iGlobalDataSize = reader.ReadInt();
    //INFO_MSG("global data size: %d",out->iGlobalDataSize);

    num = reader.ReadInt();
    //INFO_MSG("number of instructions: %d",num);

    FillInstructionTable(out,reader,num);

    num = reader.ReadInt();
    //INFO_MSG("number of strings: %d",num);

    FillStringTable(out,reader,num);

    num = reader.ReadInt();
    //INFO_MSG("number of functions: %d",num);

    FillFunctionTable(out,reader,num);

    num = reader.ReadInt();
    //INFO_MSG("number of HostAPI calls: %d",num);

    FillHostAPITable(out,reader,num);

    num = reader.ReadInt();
    //INFO_MSG("number of HostObj calls: %d",num);

    FillHostObjTable(out,reader,num);

	num = reader.ReadInt();
	FillDebugTable(out,reader,num);

    reader.CloseFile();

    //CheckHAPIProviders(out);

    //INFO_MSG("successfully loaded ISE file %s",filename.c_str());

    m_ISEStructs[filename] = out;

    return out;
}


void IsisVM::FillInstructionTable(ISEStruct * ise, ISEReader & reader, 
                                  int num_instr)
{
    for (int i = 0; i < num_instr; i++)
    {
        Instruction inst;
        inst.Load(reader);
        ise->InstructionTable.push_back(inst);
    }
}

void IsisVM::FillStringTable(ISEStruct * ise,
                             ISEReader & reader, int num_strings)
{
    for (int i = 0; i < num_strings; i++)
    {
        ise->StringTable.push_back(reader.ReadString());
    }
}

void IsisVM::FillFunctionTable(ISEStruct * ise, ISEReader & reader, 
                               int num_functions)
{
    for (int i = 0; i < num_functions; i++)
    {
        int first_instr;
        int num_params;
        int locals_size;
        std::string name;

        first_instr = reader.ReadInt();
        num_params = reader.ReadInt();
        locals_size = reader.ReadInt();
        name = reader.ReadString();

        ise->FunctionTable.push_back(Function(first_instr,num_params,
                                              locals_size,name));
    }
}

void IsisVM::FillHostAPITable(ISEStruct * ise, ISEReader & reader, 
                              int num_functions)
{
    for (int i = 0; i < num_functions; i++)
    { 
        ise->HostAPITable.push_back(reader.ReadString());
    }
}

void IsisVM::FillHostObjTable(ISEStruct * ise, ISEReader & reader, 
                              int num_functions)
{
    for (int i = 0; i < num_functions; i++)
    { 
        ise->HostObjTable.push_back(reader.ReadString());
    }
}

void IsisVM::FillDebugTable(ISEStruct * ise, ISEReader & reader, 
                              int num_items)
{
    for (int i = 0; i < num_items; i++)
    { 
		int start = reader.ReadInt();
		int end = reader.ReadInt();

		ise->DebugTable.push_back(std::pair<int,int>(start,end));
    }
}

void IsisVM::CheckHAPIProviders(ISEStruct * ise)
{
    for (unsigned int i = 0; i < ise->HostAPITable.size(); i++)
    {
        if (!HAPIProviderExists(ise->HostAPITable[i]))
            throw IVMException("host API function not found in VM");
    }
}

} // namespace IVM

