
#include <isis/util.h>

#include <isis/ivm/iseformat.h>

#include <isis/ivm/IVMException.h>
#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Script.h>

#include "ISEStruct.h"

namespace IVM {

void Script::ExecSync(const std::string & funcname,
                      const std::vector<Value> & parameters)
{
    if (!m_bLoaded)
        throw IVMException("not loaded in ExecAsync");

	AddRef();

    m_NextInstrStack.push_back(m_iNextInstr);
    m_iNextInstr = -1;

    int funcindex = FindFuncIndex(funcname);

    if (-1 == funcindex)
		throw IVMException("Unknown function:",funcname);

    if ((int)parameters.size() != m_pISE->FunctionTable[funcindex].iNumParams)
		throw IVMException("Incorrect number of parameters for:",funcname);

    for (unsigned int i = 0; i < parameters.size(); i++)
    {
        Push(parameters[i]);
    }

	try
	{
	    CallFunction(funcindex,(int)parameters.size());

		while (ExecNextInstruction());
	}
	catch (Isis::IsisException)
	{
		int line = -1;
		for (int i = 0; i < (int)m_pISE->DebugTable.size(); i++)
		{
			if (m_iCurrInstr >= m_pISE->DebugTable[i].second)
			{
				line = m_pISE->DebugTable[i].first;
//				break;
			}
		}

		ERR_MSG("in function: %s::%s\nat line %d in file %s",
				m_pParentVM->NameOfISE(m_pISE).c_str(),
				m_pISE->FunctionTable[funcindex].sName.c_str(),
				line,m_pISE->Filename.c_str());

		throw;
	}

    if (-1 != m_iNextInstr)
    {
        WARN_MSG("stack not properly cleaned up by script");
    }

    m_iNextInstr = m_NextInstrStack.back();
    m_NextInstrStack.pop_back();

	Release();
}

bool Script::ExecNextInstruction()
{
    if (!m_bLoaded)
        throw IVMException("not loaded in ExecuteNextInstruction");

    m_iCurrInstr = m_iNextInstr++;

    //INFO_MSG("Executing instr: %d",m_iCurrInstr);

    if (m_iCurrInstr < 0 || 
        m_iCurrInstr > (int)m_pISE->InstructionTable.size()-1)
    {
        throw IVMException("attempt to fetch invalid instruction");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].Type)
    {
    case INSTR_MOV:
        {
            Value value;

            SetOpAsLValue(GetOpAsRValue(1),0);
        } break;
    case INSTR_ADD:
    case INSTR_SUB:
    case INSTR_MUL:
    case INSTR_DIV:
    case INSTR_MOD:
    case INSTR_LOGICAND:
    case INSTR_LOGICOR:
    case INSTR_BITAND:
    case INSTR_BITOR:
    case INSTR_XOR:
    case INSTR_SHL:
    case INSTR_SHR:
    case INSTR_CMPE:
    case INSTR_CMPNE:
    case INSTR_CMPG:
    case INSTR_CMPGE:
    case INSTR_CMPL:
    case INSTR_CMPLE:
    case INSTR_CONCAT:
        {
            Value val1;
            Value val2;
            Value result;

            val1 = GetOpAsRValue(1);
            val2 = GetOpAsRValue(2);

            switch (m_pISE->InstructionTable[m_iCurrInstr].Type)
            {
            case INSTR_ADD:      result = val1 + val2; break;
            case INSTR_SUB:      result = val1 - val2; break;
            case INSTR_MUL:      result = val1 * val2; break;
            case INSTR_DIV:      result = val1 / val2; break;
            case INSTR_MOD:      result = val1 % val2; break;
            case INSTR_LOGICAND: result = val1 && val2; break;
            case INSTR_LOGICOR:  result = val1 || val2; break;
            case INSTR_BITAND:   result = val1 & val2; break;
            case INSTR_BITOR:    result = val1 | val2; break;
            case INSTR_XOR:      result = val1 ^ val2; break;
            case INSTR_SHL:      result = val1 << val2; break;
            case INSTR_SHR:      result = val1 >> val2; break;
            case INSTR_CMPE:     result = val1 == val2; break;
            case INSTR_CMPNE:    result = val1 != val2; break;
            case INSTR_CMPG:     result = val1 >  val2; break;
            case INSTR_CMPGE:    result = val1 >= val2; break;
            case INSTR_CMPL:     result = val1 <  val2; break;
            case INSTR_CMPLE:    result = val1 <= val2; break;
            case INSTR_CONCAT:   result = val1.GetString() + 
                                          val2.GetString(); break;
            default: break;
            }

            SetOpAsLValue(result,0);

        } break;
    case INSTR_NEG:
    case INSTR_INC:
    case INSTR_DEC:
    case INSTR_NOT:
        {
            Value value = GetOpAsRValue(1);

            switch (m_pISE->InstructionTable[m_iCurrInstr].Type)
            {
            case INSTR_NEG: value = -value; break;
            case INSTR_INC: value = value+1; break;
            case INSTR_DEC: value = value-1; break;
            case INSTR_NOT: value = !value; break;
            default: break;
            }

            SetOpAsLValue(value,0);
        } break;
    case INSTR_GETCHAR:
        {
            Value val1;
            Value val2;
            Value result;
            std::string text;
            int offset;

            val1 = GetOpAsRValue(1);
            val2 = GetOpAsRValue(2);

            text = val1.GetString();
            offset = val2.GetInteger();

            if (offset < 0) offset = 0;
            if (offset > (int)text.length()-1) offset = (int)text.length()-1;

            result = std::string(1,text[offset]);

            SetOpAsLValue(result,0);
        } break;
    case INSTR_SETCHAR:
        {
            Value val1;
            Value val2;
            Value val3;
            std::string text1;
            std::string text2;
            int offset;
            char character;

            val1 = GetOpAsRValue(0);
            val2 = GetOpAsRValue(1);
            val3 = GetOpAsRValue(2);

            text1 = val1.GetString();
            offset = val2.GetInteger();
            text2 = val3.GetString();

            if (0 == text2.length())
                character = '#';
            else
                character = text2[0];

            if (offset < 0) offset = 0;
            if (offset > (int)text1.length()-1) offset = (int)text1.length()-1;

            if (0 == text1.length())
                text1 = character;
            else
                text1[offset] = character;

            val1 = text1;

            SetOpAsLValue(val1,0);
        } break;
    case INSTR_JMP:
        {
            m_iNextInstr = GetOpAsInstrNum(0);
        } break;
    case INSTR_JZ:
    case INSTR_JNZ:
        {
            int target;
            Value value;
            bool result;

            value = GetOpAsRValue(0);
            target = GetOpAsInstrNum(1);

            switch (m_pISE->InstructionTable[m_iCurrInstr].Type)
            {
            case INSTR_JZ:  result = (0 == value); break;
            case INSTR_JNZ: result = (0 != value); break;
            default: result = false; break;
            }

            if (result)
                m_iNextInstr = target;

        } break;
    case INSTR_PUSH:
        {
            Push(GetOpAsRValue(0));
        } break;
    case INSTR_POP:
        {
            SetOpAsLValue(Pop(),0);
        } break;
    case INSTR_CALL:
        {
            CallFunction(GetOpAsFuncIndex(0),GetOpAsRValue(1));
        } break;
    case INSTR_CALLHOST:
        {
            CallHostFunction(GetOpAsHostFuncIndex(0),GetOpAsRValue(1));
        } break;
    case INSTR_CALLOBJ:
        {
            CallHostObject(GetOpAsHostObjPtr(0),
                           GetOpAsHostObjIndex(1),
                           GetOpAsRValue(2));
        } break;
    case INSTR_THIS:
        {
            SetRetVal(this);
        } break;
    case INSTR_RET:
        {
            return Return();
        } break;
    default:
        {
            throw IVMException("attempt to execute unknown instruction type");
        }
    }

    return true;
}

} // namespace IVM

