
#include <isis/util.h>

#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Script.h>
#include <isis/ivm/IVMException.h>
#include <isis/ivm/IVMVector.h>

#include "ISEStruct.h"

namespace IVM {

void Script::FillExportTable()
{
    m_ExportTable["IsLoaded"] =            FUNC_ISLOADED;
    m_ExportTable["ContainsFunction"] =    FUNC_CONTAINSFUNCTION;
    m_ExportTable["ExecNextInstruction"] = FUNC_EXECNEXTINSTRUCTION;
    m_ExportTable["SetRetVal"] =           FUNC_SETRETVAL;
    m_ExportTable["GetRetVal"] =           FUNC_GETRETVAL;
    m_ExportTable["GetParentVM"] =         FUNC_GETPARENTVM;
    m_ExportTable["DebugPrint"] =          FUNC_DEBUGPRINT;
    m_ExportTable["Call"] =                FUNC_CALL;
}

bool Script::IVMContainsMethod(const std::string & funcname)
{
    bool bScriptFunc = (m_ExportTable.end() != m_ExportTable.find(funcname));

    if (bScriptFunc || ContainsFunction(funcname))
        return true;
    else 
        return false;
}

void Script::IVMFillExportTable(std::vector<std::string> & table)
{
/*    std::map<std::string,FuncType>::iterator iter;

    for (iter = m_ExportTable.begin(); iter != m_ExportTable.end(); iter++)
    {
        table.push_back(iter->first);
    }*/

	for (unsigned int i = 0; i < m_pISE->FunctionTable.size(); i++)
	{
		table.push_back(m_pISE->FunctionTable[i].sName);
	}
}

Value Script::IVMCallMethod(const std::string & funcname, 
                            const std::vector<Value> & parameters)
{
    bool bScriptFunc;

    bScriptFunc = (m_ExportTable.end() != m_ExportTable.find(funcname));

    if (!bScriptFunc && !ContainsFunction(funcname))
		throw IVMException("function not found in Script object:",
							funcname);

    if (!bScriptFunc)
    {
        ExecSync(funcname,parameters);

        return GetRetVal();
    }

    switch (m_ExportTable[funcname])
    {
    case FUNC_ISLOADED:
        {
            if (parameters.size() > 0)
            {
                WARN_MSG("parameters to IsLoaded() ignored");
            }

            return IsLoaded();
        } break;
    case FUNC_CONTAINSFUNCTION:
        {
            if (parameters.size() == 0)
                throw IVMException(
                         "not enough parameters for ContainsFunction()");

            if (parameters.size() > 1)
            {
                WARN_MSG("extra parameters to ContainsFunction() ignored");
            }

            return ContainsFunction(parameters[0].GetString());
        } break;
    case FUNC_EXECNEXTINSTRUCTION:
        {
            if (parameters.size() > 0)
                WARN_MSG("parameters ignored for ExecNextInstruction");

            ExecNextInstruction();

            return true;

        } break;
    case FUNC_SETRETVAL:
        {
            if (parameters.size() == 0)
                throw IVMException("not enough parameters for SetRetVal()");

            if (parameters.size() > 1)
            {
                WARN_MSG("extra parameters for SetRetVal() ignored");
            }

            SetRetVal(parameters[0]);

            return true;

        } break;
    case FUNC_GETRETVAL:
        {
            if (parameters.size() > 0)
            {
                WARN_MSG("extra parameters for GetRetVal() ignored");
            }

            return GetRetVal();

        } break;
    case FUNC_GETPARENTVM:
        {
            if (parameters.size() >= 0)
            {
                WARN_MSG("extra parameters for GetParentVM ignored");
            }

            return GetParentVM();

        } break;
    case FUNC_DEBUGPRINT:
        {
            if (parameters.size() > 0)
            {
                WARN_MSG("extra parameters for DebugPrint() ignored");
            }

            DebugPrint();

            return true;

        } break;
    case FUNC_CALL:
        {
			std::vector<Value> params;

			switch (parameters.size())
			{
			case 1:
				{
					return IVMCallMethod(parameters[0].GetString(),params);
				} break;
			case 2:
				{
					IVMVector * pVector = dynamic_cast<IVMVector*>(
											parameters[1].GetObjPtr());

					if (pVector)
					{
						for (int i = 0; i < pVector->Size(); i++)
							params.push_back(pVector->Get(i));
					}

					return IVMCallMethod(parameters[0].GetString(),params);

				} break;
			default:
				{
					throw IVMException(
						"wrong number of parameters for Call()");
				} break;
			}
        } break;
    }

	return true;
}

} // namespace IVM

