
#include <algorithm>

#include <isis/util.h>

#include <isis/ivm/IVMException.h>
#include <isis/ivm/Script.h>
#include <isis/ivm/IsisVM.h>
#include <isis/ivm/Value.h>

#include "ISEStruct.h"

namespace IVM {

void Script::Push(const Value & value)
{
    m_Stack.push_back(value);
}

Value Script::Pop()
{
    Value value;

    if (0 == m_Stack.size())
        throw IVMException("Stack underflow");

    value = m_Stack.back();

    m_Stack.pop_back();

    return value;
}

void Script::CallFunction(int funcindex, const Value & num_params)
{
    if (funcindex < 0 || funcindex > (int)m_pISE->FunctionTable.size()-1)
        throw IVMException("invalid function index in CallFunction");

    if (num_params.GetInteger() != m_pISE->FunctionTable[funcindex].iNumParams)
        throw IVMException("incorrect number of parameters passed to function");
    
    Push(m_iNextInstr);

    for (int i = 0; i < m_pISE->FunctionTable[funcindex].iLocalsSize; i++)
    {
        Push(Value());
    }

    Push(m_iCurrStackFrame);
    Push(funcindex);

    m_iCurrStackFrame = (int)m_Stack.size();
    m_iNextInstr = m_pISE->FunctionTable[funcindex].iFirstInstr;
}

void Script::CallHostFunction(int hostfuncindex, const Value & num_params)
{
    IVMObject * provider;
    Value v;
    std::vector<Value> params;
    int i;

    if (hostfuncindex < 0 || hostfuncindex > (int)m_pISE->HostAPITable.size()-1)
        throw("host api func index is invalid");

    provider = m_pParentVM->FindHAPIProvider(
                                     m_pISE->HostAPITable[hostfuncindex]);

    if (NULL == provider)
    {
		throw IVMException("error finding host API func:",
							m_pISE->HostAPITable[hostfuncindex]);
    }

    for (i = num_params.GetInteger() - 1; i >= 0; i--)
    {
        int num = (int)m_Stack.size()-1-i;

        if (num < 0 || num > (int)m_Stack.size()-1)
        {
            throw IVMException(
                   "invalid stack index for parameters in CallHostFunction");
        }
        
        params.push_back(m_Stack[num]);
    }

    SetRetVal(provider->IVMCallMethod(m_pISE->HostAPITable[hostfuncindex],params));

    for (i = 0; i < num_params.GetInteger(); i++)
    {
        Pop();
    }
}

void Script::CallHostObject(IVMObject * object, int hostobjfunc,
                            const Value & num_params)
{
    Value v;
    std::vector<Value> params;
    int i;

    if (hostobjfunc < 0 || hostobjfunc > (int)m_pISE->HostObjTable.size()-1)
        throw IVMException("host api func index is invalid");

    if (NULL == object)
		throw IVMException("NULL pointer passed as object to call function:",
						m_pISE->HostObjTable[hostobjfunc]);

	for (i = num_params.GetInteger() - 1; i >= 0; i--)
	{
		int num = (int)m_Stack.size()-1-i;

		if (num < 0 || num > (int)m_Stack.size()-1)
			throw IVMException( 
				"invalid stack index for parameters in CallHostObject");

		params.push_back(m_Stack[num]);
	}

    SetRetVal(object->IVMCallMethod(m_pISE->HostObjTable[hostobjfunc],params));

    for (i = 0; i < num_params.GetInteger(); i++)
    {
        Pop();
    }
}

bool Script::Return()
{
    Value funcindex;
    Value stackframe;
    Value returnaddr;

    funcindex = Pop();
    stackframe = Pop();

    for (int i = 0; 
         i < m_pISE->FunctionTable[funcindex.GetInteger()].iLocalsSize; 
         i++)
    {
        Pop();
    }

    returnaddr = Pop();

    for (int i = 0;
         i < m_pISE->FunctionTable[funcindex.GetInteger()].iNumParams;
         i++)
    {
        Pop();
    }

    m_iCurrStackFrame = stackframe.GetInteger();
    m_iNextInstr = returnaddr.GetInteger();

    if (-1 == m_iNextInstr)
        return false;

    return true;
}

Value Script::GetOpAsRValue(int opnum)
{
    Value value;

    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_STRING:
        {
            value.AssignString(m_pISE->StringTable[m_pISE->InstructionTable[
                              m_iCurrInstr].OpList[opnum].iStringIndex]);
        } break;
    case OP_INTEGER:
        {
            value.AssignInteger(
                  m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iValue);
        } break;
    case OP_FLOAT:
        {
            value.AssignFloat(
                  m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].fValue);
        } break;
    case OP_RETVAL:
        {
            value = m_RetVal;
        } break;
    case OP_SCALAR:
        {
            int stackind = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind < 0)
                stackind += m_iCurrStackFrame;

            if (stackind < 0 || stackind > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            value = m_Stack[stackind];

        } break;
    case OP_ARRAY:
        {
            int stackind1;
            int stackind2;
            
            stackind1 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind1 < 0)
                stackind1 += m_iCurrStackFrame;

            if (stackind1 < 0 || stackind1 > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            stackind2 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iExtraInfo;

            if (stackind2 < 0)
                stackind2 += m_iCurrStackFrame;

            if (stackind2 < 0 || stackind2 > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            value = m_Stack[stackind1 + m_Stack[stackind2].GetInteger()];

        } break;
    default:
        {
            throw IVMException("operand not an rvalue type");
        } break;
    }

    return value;
}

void Script::SetOpAsLValue(const Value & value, int opnum)
{
    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_RETVAL:
        {
            m_RetVal = value;
        } break;
    case OP_SCALAR:
        {
            int stackind = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind < 0)
                stackind += m_iCurrStackFrame;

            if (stackind < 0 || stackind > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            m_Stack[stackind] = value;

        } break;
    case OP_ARRAY:
        {
            int stackind1;
            int stackind2;
            
            stackind1 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind1 < 0)
                stackind1 += m_iCurrStackFrame;

            if (stackind1 < 0 || stackind1 > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            stackind2 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iExtraInfo;

            if (stackind2 < 0)
                stackind2 += m_iCurrStackFrame;

            if (stackind2 < 0 || stackind2 > (int)m_Stack.size()-1)
            {
                throw IVMException("invalid stack index");
            }

            m_Stack[stackind1 + m_Stack[stackind2].GetInteger()] = value;

        } break;
    default:
        {
            throw IVMException("operand type not an lvalue type");
        } break;
    }
}

int Script::GetOpAsInstrNum(int opnum)
{
    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_LABEL:
        {
            return
                m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iInstrNum;
        } break;
    default:
        {
            throw IVMException("operand type not an instrnum");
        } break;
    }
}

int Script::GetOpAsFuncIndex(int opnum)
{
    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_FUNCINDEX:
        {
            return
               m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iFuncIndex;
        } break;
    default:
        {
            throw IVMException("operand type not a funcindex");
        } break;
    }
}

int Script::GetOpAsHostFuncIndex(int opnum)
{
    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_HOSTFUNC:
        {
            return
           m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iHostFuncIndex;
        } break;
    default:
        {
            throw IVMException("operand type not a hostfuncindex");
        } break;
    }
}

int Script::GetOpAsHostObjIndex(int opnum)
{
    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_HOSTOBJ:
        {
            return 
           m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iHostObjIndex;
        } break;
    default:
        {
            throw IVMException("operand type not a hostfuncindex");
        } break;
    }
}

IVMObject * Script::GetOpAsHostObjPtr(int opnum)
{
    Value value;

    if (opnum < 0 || 
        opnum > (int)m_pISE->InstructionTable[m_iCurrInstr].OpList.size()-1)
    {
        throw IVMException("invalid operand number");
    }

    switch (m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].Type)
    {
    case OP_RETVAL:
        {
            value = m_RetVal;
        } break;
    case OP_SCALAR:
        {
            int stackind = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind < 0)
                stackind += m_iCurrStackFrame;

            if (stackind < 0 || stackind > (int)m_Stack.size()-1)
                throw IVMException("invalid stack index");

            value = m_Stack[stackind];

        } break;
    case OP_ARRAY:
        {
            int stackind1;
            int stackind2;
            
            stackind1 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iStackIndex;

            if (stackind1 < 0)
                stackind1 += m_iCurrStackFrame;

            if (stackind1 < 0 || stackind1 > (int)m_Stack.size()-1)
                throw IVMException("invalid stack index");

            stackind2 = 
             m_pISE->InstructionTable[m_iCurrInstr].OpList[opnum].iExtraInfo;

            if (stackind2 < 0)
                stackind2 += m_iCurrStackFrame;

            if (stackind2 < 0 || stackind2 > (int)m_Stack.size()-1)
                throw IVMException("invalid stack index");

            value = m_Stack[stackind1 + m_Stack[stackind2].GetInteger()];

        } break;
    default:
        {
            throw IVMException("operand type not an object type");
        } break;
    }

    return value.GetObjPtr();
}


} // namespace IVM

